import os
import logging
import logging.handlers

import gdrivefs.config

logger = logging.getLogger()

if gdrivefs.config.IS_DEBUG:
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.WARNING)

def _configure_syslog():
    facility = logging.handlers.SysLogHandler.LOG_LOCAL0
    sh = logging.handlers.SysLogHandler(facility=facility)
    formatter = logging.Formatter('GD: %(name)-12s %(levelname)-7s %(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)

def _configure_file():
    filepath = os.environ.get('GD_LOG_FILEPATH', '/tmp/gdrivefs.log')
    fh = logging.FileHandler(filepath)
    formatter = logging.Formatter('%(asctime)s [%(name)s %(levelname)s] %(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

def _configure_console():
    sh = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s [%(name)s %(levelname)s] %(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)

if gdrivefs.config.IS_DEBUG is True:
#    _configure_file()
    _configure_console()
else:
    _configure_syslog()



import logging

from collections    import deque
from threading      import RLock
from datetime       import datetime

from gdrivefs.utility import utility
from gdrivefs.conf import Conf
from gdrivefs.gdtool.drive import get_gdrive
from gdrivefs.gdtool.account_info import AccountInfo
from gdrivefs.gdtool.normal_entry import NormalEntry
from gdrivefs.cache.cache_registry import CacheRegistry, CacheFault
from gdrivefs.cache.cacheclient_base import CacheClientBase
from gdrivefs.errors import GdNotFoundError

CLAUSE_ENTRY            = 0 # Normalized entry.
CLAUSE_PARENT           = 1 # List of parent clauses.
CLAUSE_CHILDREN         = 2 # List of 2-tuples describing children: (filename, clause)
CLAUSE_ID               = 3 # Entry ID.
CLAUSE_CHILDREN_LOADED  = 4 # All children loaded?

_logger = logging.getLogger(__name__)

def path_resolver(path):
    path_relations = PathRelations.get_instance()

    parent_clause = path_relations.get_clause_from_path(path)
    if not parent_clause:
#        logging.debug("Path [%s] does not exist for split.", path)
        raise GdNotFoundError()

    return (parent_clause[CLAUSE_ENTRY], parent_clause)


class PathRelations(object):
    """Manages physical path representations of all of the entries in our "
    account.
    """

    rlock = RLock()

    entry_ll = { }
    path_cache = { }
    path_cache_byid = { }

    @staticmethod
    def get_instance():

        with PathRelations.rlock:
            try:
                return CacheRegistry.__instance;
            except:
                pass

            CacheRegistry.__instance = PathRelations()
            return CacheRegistry.__instance

    def remove_entry_recursive(self, entry_id, is_update=False):
        """Remove an entry, all children, and any newly orphaned parents."""

        _logger.debug("Recursively pruning entry with ID [%s].", entry_id)

        to_remove = deque([ entry_id ])
        stat_placeholders = 0
        stat_folders = 0
        stat_files = 0
        removed = { }
        while 1:
            if not to_remove:
                break

            current_entry_id = to_remove.popleft()
            entry_clause = self.entry_ll[current_entry_id]

            # Any entry that still has children will be transformed into a 
            # placeholder, and not actually removed. Once the children are 
            # removed in this recursive process, we'll naturally clean-up the 
            # parent as a last step. Therefore, the number of placeholders will 
            # overlap with the number of folders (a placeholder must represent 
            # a folder. It is only there because the entry had children).

            if not entry_clause[0]:
                stat_placeholders += 1
            elif entry_clause[0].is_directory:
                stat_folders += 1
            else:
                stat_files += 1

            result = self.__remove_entry(current_entry_id, is_update)

            removed[current_entry_id] = True

            (current_orphan_ids, current_children_clauses) = result

            children_ids_to_remove = [ children[3] for children 
                                                in current_children_clauses ]

            to_remove.extend(current_orphan_ids)
            to_remove.extend(children_ids_to_remove)

        return (removed.keys(), (stat_folders + stat_files))

    def __remove_entry(self, entry_id, is_update=False):
        """Remove an entry. Updates references from linked entries, but does 
        not remove any other entries. We return a tuple, where the first item 
        is a list of any parents that, themselves, no longer have parents or 
        children, and the second item is a list of children to this entry.
        """

        with PathRelations.rlock:
            # Ensure that the entry-ID is valid.

            entry_clause = self.entry_ll[entry_id]
            
            # Clip from path cache.

            if entry_id in self.path_cache_byid:
                path = self.path_cache_byid[entry_id]
                del self.path_cache[path]
                del self.path_cache_byid[entry_id]

            # Clip us from the list of children on each of our parents.

            entry_parents = entry_clause[CLAUSE_PARENT]
            entry_children_tuples = entry_clause[CLAUSE_CHILDREN]

            parents_to_remove = [ ]
            children_to_remove = [ ]
            if entry_parents:
                for parent_clause in entry_parents:
                    # A placeholder has an entry and parents field (fields 
                    # 0, 1) of None.

                    (parent, parent_parents, parent_children, parent_id, \
                        all_children_loaded) = parent_clause

                    if all_children_loaded and not is_update:
                        all_children_loaded = False

                    # Integrity-check that the parent we're referencing is 
                    # still in the list.
                    if parent_id not in self.entry_ll:
                        _logger.warn("Parent with ID [%s] on entry with ID "
                                     "[%s] is not valid." % 
                                     (parent_id, entry_id))
                        continue
            
                    old_children_filenames = [ child_tuple[0] for child_tuple 
                                                in parent_children ]

                    updated_children = [ child_tuple for child_tuple 
                                         in parent_children 
                                         if child_tuple[1] != entry_clause ]

                    if parent_children != updated_children:
                        parent_children[:] = updated_children

                    else:
                        _logger.error("Entry with ID [%s] referenced parent "
                                      "with ID [%s], but not vice-versa." % 
                                      (entry_id, parent_id))

                    updated_children_filenames = [ child_tuple[0] 
                                                    for child_tuple
                                                    in parent_children ]

                    # If the parent now has no children and is a placeholder, 
                    # advise that we remove it.
                    if not parent_children and parent == None:
                        parents_to_remove.append(parent_id)

            # Remove/neutralize entry, now that references have been removed.

            set_placeholder = len(entry_children_tuples) > 0

            if set_placeholder:
                # Just nullify the entry information, but leave the clause. We 
                # had children that still need a parent.

                entry_clause[0] = None
                entry_clause[1] = None
            else:
                del self.entry_ll[entry_id]

        children_entry_clauses = [ child_tuple[1] for child_tuple 
                                    in entry_children_tuples ]

        return (parents_to_remove, children_entry_clauses)

    def remove_entry_all(self, entry_id, is_update=False):
        """Remove the the entry from both caches. EntryCache is more of an 
        entity look-up, whereas this (PathRelations) has a bunch of expanded 
        data regarding relationships and paths. This call will first remove the 
        relationships from here, and then the entry from the EntryCache.

        We do it in this order because if we were to remove entry from the core
        library (EntryCache) first, then all of the relationships here will 
        suddenly become invalid, and although the entry will be disregistered,
        because it has references from this linked-list, those objects will be
        very much alive. On the other hand, if we remove the entry from 
        PathRelations first, then, because of the locks, PathRelations will not
        be able to touch the relationships until after we're done, here. Ergo, 
        the only thing that can happen is that something may look at the entry
        in the library.
        """

        with PathRelations.rlock:
            cache = EntryCache.get_instance().cache

            removed_ids = [ entry_id ]
            if self.is_cached(entry_id):
                try:
                    removed_tuple = self.remove_entry_recursive(entry_id, \
                                                               is_update)
                except:
                    _logger.exception("Could not remove entry-ID from "
                                      "PathRelations. Still continuing, "
                                      "though.")

                (removed_ids, number_removed) = removed_tuple

            for removed_id in removed_ids:
                if cache.exists(removed_id):
                    try:
                        cache.remove(removed_id)
                    except:
                        _logger.exception("Could not remove entry-ID from "
                                          "the core cache. Still "
                                          "continuing, though.")

    def get_proper_filenames(self, entry_clause):
        """Return what was determined to be the unique filename for this "
        particular entry for each of its respective parents. This will return 
        the standard 'title' value as a scalar when the root entry, and a 
        dictionary of parent-IDs to unique-filenames when not.

        This call is necessary because GD allows duplicate filenames until any 
        one folder. Note that a consequence of both this and the fact that GD 
        allows the same file to be listed under multiple folders means that a 
        file may look like "filename" under one and "filename (2)" under 
        another.
        """

        with PathRelations.rlock:
            found = { }
            parents = entry_clause[1]
            if not parents:
                return entry_clause[0].title_fs

            else:
                for parent_clause in parents:
                    matching_children = [filename for filename, child_clause 
                                                  in parent_clause[2] 
                                                  if child_clause == entry_clause]
                    if not matching_children:
                        _logger.error("No matching entry-ID [%s] was not "
                                      "found among children of entry's "
                                      "parent with ID [%s] for proper-"
                                      "filename lookup." % 
                                      (entry_clause[3], parent_clause[3]))

                    else:
                        found[parent_clause[3]] = matching_children[0]

        return found

    def register_entry(self, normalized_entry):

        with PathRelations.rlock:
            if not normalized_entry.is_visible:
                return None

            if normalized_entry.__class__ is not NormalEntry:
                raise Exception("PathRelations expects to register an object "
                                "of type NormalEntry, not [%s]." % 
                                (type(normalized_entry)))

            entry_id = normalized_entry.id

#            self.__log.debug("Registering entry with ID [%s] within path-"
#                             "relations.", entry_id)

            if self.is_cached(entry_id, include_placeholders=False):
                self.remove_entry_recursive(entry_id, True)

            cache = EntryCache.get_instance().cache

            cache.set(normalized_entry.id, normalized_entry)

            # We do a linked list using object references.
            # (
            #   normalized_entry, 
            #   [ parent clause, ... ], 
            #   [ child clause, ... ], 
            #   entry-ID,
            #   < boolean indicating that we know about all children >
            # )

            if self.is_cached(entry_id, include_placeholders=True):
                entry_clause = self.entry_ll[entry_id]
                entry_clause[CLAUSE_ENTRY] = normalized_entry
                entry_clause[CLAUSE_PARENT] = [ ]
            else:
                entry_clause = [normalized_entry, [ ], [ ], entry_id, False]
                self.entry_ll[entry_id] = entry_clause

            entry_parents = entry_clause[CLAUSE_PARENT]
            title_fs = normalized_entry.title_fs

            parent_ids = normalized_entry.parents if normalized_entry.parents \
                                                  is not None else []

            for parent_id in parent_ids:

                # If the parent hasn't yet been loaded, install a placeholder.
                if self.is_cached(parent_id, include_placeholders=True):
                    parent_clause = self.entry_ll[parent_id]
                else:
                    parent_clause = [None, None, [ ], parent_id, False]
                    self.entry_ll[parent_id] = parent_clause

                if parent_clause not in entry_parents:
                    entry_parents.append(parent_clause)

                parent_children = parent_clause[CLAUSE_CHILDREN]
                filename_base = title_fs

                # Register among the children of this parent, but make sure we 
                # have a unique filename among siblings.

                i = 0
                current_variation = filename_base
                elected_variation = None
                while i <= 255:
                    if not [ child_name_tuple 
                             for child_name_tuple 
                             in parent_children 
                             if child_name_tuple[0] == current_variation ]:
                        elected_variation = current_variation
                        break
                        
                    i += 1
                    current_variation = filename_base + \
                                        utility.translate_filename_charset(
                                            ' (%d)' % (i))

                if elected_variation == None:
                    _logger.error("Could not register entry with ID [%s]. "
                                  "There are too many duplicate names in "
                                  "that directory." % (entry_id))
                    return

                # Register us in the list of children on this parents 
                # child-tuple list.
                parent_children.append((elected_variation, entry_clause))

        return entry_clause

    def __load_all_children(self, parent_id):
        gd = get_gdrive()

        with PathRelations.rlock:
            children = gd.list_files(parent_id=parent_id)

            child_ids = [ ]
            if children:
                for child in children:
                        self.register_entry(child)

                parent_clause = self.__get_entry_clause_by_id(parent_id)

                parent_clause[4] = True

        return children

    def get_children_from_entry_id(self, entry_id):
        """Return the filenames contained in the folder with the given 
        entry-ID.
        """

        with PathRelations.rlock:
            entry_clause = self.__get_entry_clause_by_id(entry_id)
            if not entry_clause:
                message = ("Can not list the children for an unavailable "
                           "entry with ID [%s]." % (entry_id))

                _logger.error(message)
                raise Exception(message)

            if not entry_clause[4]:
                self.__load_all_children(entry_id)

            if not entry_clause[0].is_directory:
                message = ("Could not get child filenames for non-directory with "
                           "entry-ID [%s]." % (entry_id))

                _logger.error(message)
                raise Exception(message)

#            self.__log.debug("(%d) children found.",
#                             len(entry_clause[CLAUSE_CHILDREN]))

            return entry_clause[CLAUSE_CHILDREN]

    def get_children_entries_from_entry_id(self, entry_id):

        children_tuples = self.get_children_from_entry_id(entry_id)

        children_entries = [(child_tuple[0], child_tuple[1][CLAUSE_ENTRY]) 
                                for child_tuple 
                                in children_tuples]

        return children_entries

    def get_clause_from_path(self, filepath):

#        self.__log.debug("Getting clause for path [%s].", filepath)

        with PathRelations.rlock:
            path_results = self.find_path_components_goandget(filepath)

            (entry_ids, path_parts, success) = path_results
            if not success:
                return None

            entry_id = path_results[0][-1]
#            self.__log.debug("Found entry with ID [%s].", entry_id)

            # Make sure the entry is more than a placeholder.
            self.__get_entry_clause_by_id(entry_id)

            return self.entry_ll[entry_id]

    def find_path_components_goandget(self, path):
        """Do the same thing that find_path_components() does, except that 
        when we don't have record of a path-component, try to go and find it 
        among the children of the previous path component, and then try again.
        """

        gd = get_gdrive()

        with PathRelations.rlock:
            previous_results = []
            i = 0
            while 1:
#                self.__log.debug("Attempting to find path-components (go and "
#                                 "get) for path [%s].  CYCLE= (%d)", path, i)

                # See how many components can be found in our current cache.

                result = self.__find_path_components(path)

                # If we could resolve the entire path, return success.

                if result[2] == True:
                    return result

                # If we could not resolve the entire path, and we're no more 
                # successful than a prior attempt, we'll just have to return a 
                # partial.

                num_results = len(result[0])
                if num_results in previous_results:
                    return result

                previous_results.append(num_results)

                # Else, we've encountered a component/depth of the path that we 
                # don't currently know about.
# TODO: This is going to be the general area that we'd have to adjust to 
#        support multiple, identical entries. This currently only considers the 
#        first result. We should rewrite this to be recursive in order to make 
#        it easier to keep track of a list of results.
                # The parent is the last one found, or the root if none.
                parent_id = result[0][num_results - 1] \
                                if num_results \
                                else AccountInfo.get_instance().root_id

                # The child will be the first part that was not found.
                child_name = result[1][num_results]

                children = gd.list_files(
                                parent_id=parent_id, 
                                query_is_string=child_name)
                
                for child in children:
                    self.register_entry(child)

                filenames_phrase = ', '.join([ candidate.id for candidate
                                                            in children ])
#                self.__log.debug("(%d) candidate children were found: %s",
#                                 len(children), filenames_phrase)

                i += 1

    def __find_path_components(self, path):
        """Given a path, return a list of all Google Drive entries that 
        comprise each component, or as many as can be found. As we've ensured 
        that all sibling filenames are unique, there can not be multiple 
        matches.
        """

        if path[0] == '/':
            path = path[1:]

        if len(path) and path[-1] == '/':
            path = path[:-1]

        if path in self.path_cache:
            return self.path_cache[path]

        with PathRelations.rlock:
#            self.__log.debug("Locating entry information for path [%s].", path)

            root_id = AccountInfo.get_instance().root_id

            # Ensure that the root node is loaded.
            self.__get_entry_clause_by_id(root_id)

            path_parts = path.split('/')

            entry_ptr = root_id
            parent_id = None
            i = 0
            num_parts = len(path_parts)
            results = [ ]
            while i < num_parts:
                child_filename_to_search_fs = utility. \
                    translate_filename_charset(path_parts[i])

#                self.__log.debug("Checking for part (%d) [%s] under parent "
#                                 "with ID [%s].",
#                                 i, child_filename_to_search_fs, entry_ptr)

                current_clause = self.entry_ll[entry_ptr]
            
                # Search this entry's children for the next filename further down 
                # in the path among this entry's children. Any duplicates should've 
                # already beeen handled as entries were stored. We name the variable 
                # just to emphasize that no ambiguity -as well as- no error will 
                # occur in the traversal process.
                first_matching_child_clause = None
                children = current_clause[2]
            
                # If they just wanted the "" path (root), return the root-ID.
                if path == "":
                    found = [ root_id ]
                else:
                    found = [ child_tuple[1][3] 
                              for child_tuple 
                              in children 
                              if child_tuple[0] == child_filename_to_search_fs ]

                if found:
                    results.append(found[0])
                else:
                    return (results, path_parts, False)

                # Have we traveled far enough into the linked list?
                if (i + 1) >= num_parts:
                    self.path_cache[path] = (results, path_parts, True)
                    final_entry_id = results[-1]
                    self.path_cache_byid[final_entry_id] = path

                    return self.path_cache[path]

                parent_id = entry_ptr
                entry_ptr = found[0]
                i += 1

    def __get_entry_clause_by_id(self, entry_id):
        """We may keep a linked-list of GD entries, but what we have may just 
        be placeholders. This function will make sure the data is actually here.
        """

        with PathRelations.rlock:
            if self.is_cached(entry_id):
                return self.entry_ll[entry_id]

            else:
                cache = EntryCache.get_instance().cache
                normalized_entry = cache.get(entry_id)
                return self.register_entry(normalized_entry)

    def is_cached(self, entry_id, include_placeholders=False):

        return (entry_id in self.entry_ll and (include_placeholders or \
                                               self.entry_ll[entry_id][0]))

class EntryCache(CacheClientBase):
    """Manages our knowledge of file entries."""

    def __init__(self, *args, **kwargs):
        super(EntryCache, self).__init__(*args, **kwargs)

# TODO(dustin): This isn't used, and we don't think that it necessarily needs 
#               to be instantiated, now.
#        about = AccountInfo.get_instance()
        self.__gd = get_gdrive()

    def __get_entries_to_update(self, requested_entry_id):
        # Get more entries than just what was requested, while we're at it.

        parent_ids = self.__gd.get_parents_containing_id(requested_entry_id)

        affected_entries = [requested_entry_id]
        considered_entries = {}
        max_readahead_entries = Conf.get('max_readahead_entries')
        for parent_id in parent_ids:
            child_ids = self.__gd.get_children_under_parent_id(parent_id)

            for child_id in child_ids:
                if child_id == requested_entry_id:
                    continue

                # We've already looked into this entry.

                try:
                    considered_entries[child_id]
                    continue
                except:
                    pass

                considered_entries[child_id] = True

                # Is it already cached?

                if self.cache.exists(child_id):
                    continue

                affected_entries.append(child_id)

                if len(affected_entries) >= max_readahead_entries:
                    break

        return affected_entries

    def __do_update_for_missing_entry(self, requested_entry_id):

        # Get the entries to update.

        affected_entries = self.__get_entries_to_update(requested_entry_id)

        # Read the entries, now.

# TODO: We have to determine when this is called, and either remove it 
# (if it's not), or find another way to not have to load them 
# individually.

        retrieved = self.__gd.get_entries(affected_entries)

        # Update the cache.

        path_relations = PathRelations.get_instance()

        for entry_id, entry in retrieved.iteritems():
            path_relations.register_entry(entry)

        return retrieved

    def fault_handler(self, resource_name, requested_entry_id):
        """A requested entry wasn't stored."""

        retrieved = self.__do_update_for_missing_entry(requested_entry_id)

        # Return the requested entry.
        return retrieved[requested_entry_id]

    def cleanup_pretrigger(self, resource_name, entry_id, force):
        """The core entry cache has a clean-up process that will remove old "
        entries. This is called just before any record is removed.
        """

        # Now that the local cache-item has been removed, remove the same from
        # the PathRelations cache.

        path_relations = PathRelations.get_instance()

        if path_relations.is_cached(entry_id):
            path_relations.remove_entry_recursive(entry_id)

    def get_max_cache_age_seconds(self):
        return Conf.get('cache_entries_max_age')




import logging
import re
import dateutil.parser
import random
import json
import time
import httplib
import ssl
import tempfile
import pprint
import functools
import threading
import os

import httplib2

import apiclient.discovery
import apiclient.http
import apiclient.errors

import gdrivefs.config
import gdrivefs.conf
import gdrivefs.gdtool.chunked_download
import gdrivefs.errors
import gdrivefs.gdtool.oauth_authorize
import gdrivefs.gdtool.normal_entry
import gdrivefs.time_support
import gdrivefs.gdfs.fsutility

_CONF_SERVICE_NAME = 'drive'
_CONF_SERVICE_VERSION = 'v2'

_MAX_EMPTY_CHUNKS = 3
_DEFAULT_UPLOAD_CHUNK_SIZE_B = 1024 * 1024

logging.getLogger('apiclient.discovery').setLevel(logging.WARNING)

_logger = logging.getLogger(__name__)

def _marshall(f):
    """A method wrapper that will reauth and/or reattempt where reasonable.
    """

    auto_refresh = True

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        # Now, try to invoke the mechanism. If we succeed, return 
        # immediately. If we get an authorization-fault (a resolvable 
        # authorization problem), fall through and attempt to fix it. Allow 
        # any other error to bubble up.
        
        for n in range(0, 5):
            try:
                return f(*args, **kwargs)
            except (ssl.SSLError, httplib.BadStatusLine) as e:
                # These happen sporadically. Use backoff.
                _logger.exception("There was a transient connection "
                                  "error (%s). Trying again [%s]: %s",
                                  e.__class__.__name__, str(e), n)

                time.sleep((2 ** n) + random.randint(0, 1000) / 1000)
            except apiclient.errors.HttpError as e:
                if e.content == '':
                    raise

                try:
                    error = json.loads(e.content)
                except ValueError:
                    _logger.error("Non-JSON error while doing chunked "
                                  "download: [%s]", e.content) 
                    raise e

                if error.get('code') == 403 and \
                   error.get('errors')[0].get('reason') \
                        in ['rateLimitExceeded', 'userRateLimitExceeded']:
                    # Apply exponential backoff.
                    _logger.exception("There was a transient HTTP "
                                      "error (%s). Trying again (%d): "
                                      "%s",
                                      e.__class__.__name__, str(e), n)

                    time.sleep((2 ** n) + random.randint(0, 1000) / 1000)
                else:
                    # Other error, re-raise.
                    raise
            except gdrivefs.errors.AuthorizationFaultError:
                # If we're not allowed to refresh the token, or we've
                # already done it in the last attempt.
                if not auto_refresh or n == 1:
                    raise

                # We had a resolvable authorization problem.

                _logger.info("There was an authorization fault under "
                             "action [%s]. Attempting refresh.", action)
                
                authorize = gdrivefs.gdtool.oauth_authorize.get_auth()
                authorize.check_credential_state()

                # Re-attempt the action.

                _logger.info("Refresh seemed successful. Reattempting "
                             "action [%s].", action)

    return wrapper


class GdriveAuth(object):
    def __init__(self):
        self.__client = None
        self.__authorize = gdrivefs.gdtool.oauth_authorize.get_auth()
        self.__check_authorization()
        self.__http = None

    def __check_authorization(self):
        self.__credentials = self.__authorize.get_credentials()

    def get_authed_http(self):
        if self.__http is None:
            self.__check_authorization()
            _logger.debug("Getting authorized HTTP tunnel.")
                
            http = httplib2.Http()
            self.__credentials.authorize(http)

            _logger.debug("Got authorized tunnel.")

            self.__http = http

        return self.__http

    def get_client(self):
        if self.__client is None:
            authed_http = self.get_authed_http()
        
            # Build a client from the passed discovery document path
            
            discoveryUrl = \
                gdrivefs.conf.Conf.get('google_discovery_service_url')
# TODO: We should cache this, since we have, so often, had a problem 
#       retrieving it. If there's no other way, grab it directly, and then pass
#       via a file:// URI.
        
            try:
                client = \
                    apiclient.discovery.build(
                        _CONF_SERVICE_NAME, 
                        _CONF_SERVICE_VERSION, 
                        http=authed_http, 
                        discoveryServiceUrl=discoveryUrl)
            except apiclient.errors.HttpError as e:
                # We've seen situations where the discovery URL's server is down,
                # with an alternate one to be used.
                #
                # An error here shouldn't leave GDFS in an unstable state (the 
                # current command should just fail). Hoepfully, the failure is 
                # momentary, and the next command succeeds.

                _logger.exception("There was an HTTP response-code of (%d) while "
                                  "building the client with discovery URL [%s].",
                                  e.resp.status, discoveryUrl)
                raise

            self.__client = client

        return self.__client


class _GdriveManager(object):
    """Handles all basic communication with Google Drive. All methods should
    try to invoke only one call, or make sure they handle authentication 
    refreshing when necessary.
    """

    def __init__(self):
        self.__auth = GdriveAuth()

    def __assert_response_kind(self, response, expected_kind):
        actual_kind = response[u'kind']
        if actual_kind != unicode(expected_kind):
            raise ValueError("Received response of type [%s] instead of "
                             "[%s]." % (actual_kind, expected_kind))

    @_marshall
    def get_about_info(self):
        """Return the 'about' information for the drive."""

        client = self.__auth.get_client()
        response = client.about().get().execute()
        self.__assert_response_kind(response, 'drive#about')

        return response

    @_marshall
    def list_changes(self, start_change_id=None, page_token=None):
        """Get a list of the most recent changes from GD, with the earliest 
        changes first. This only returns one page at a time. start_change_id 
        doesn't have to be valid.. It's just the lower limit to what you want 
        back. Change-IDs are integers, but are not necessarily sequential.
        """

        client = self.__auth.get_client()

        response = client.changes().list(
                    pageToken=page_token, 
                    startChangeId=start_change_id).execute()

        self.__assert_response_kind(response, 'drive#changeList')

        items = response[u'items']

        if items:
            _logger.debug("We received (%d) changes to apply.", len(items))

        largest_change_id = int(response[u'largestChangeId'])
        next_page_token = response.get(u'nextPageToken')

        changes = []
        last_change_id = None
        for item in items:
            change_id = int(item[u'id'])
            entry_id = item[u'fileId']

            if item[u'deleted']:
                was_deleted = True
                entry = None

                _logger.debug("CHANGE: [%s] (DELETED)", entry_id)
            else:
                was_deleted = False
                entry = item[u'file']

                _logger.debug("CHANGE: [%s] [%s] (UPDATED)", 
                              entry_id, entry[u'title'])

            if was_deleted:
                normalized_entry = None
            else:
                normalized_entry = \
                    gdrivefs.gdtool.normal_entry.NormalEntry(
                        'list_changes', 
                        entry)

            changes.append((change_id, (entry_id, was_deleted, normalized_entry)))
            last_change_id = change_id

        return (largest_change_id, next_page_token, changes)

    @_marshall
    def get_parents_containing_id(self, child_id, max_results=None):
        
        _logger.info("Getting client for parent-listing.")

        client = self.__auth.get_client()

        _logger.info("Listing entries over child with ID [%s].", child_id)

        response = client.parents().list(fileId=child_id).execute()
        self.__assert_response_kind(response, 'drive#parentList')

        return [ entry[u'id'] for entry in response[u'items'] ]

    @_marshall
    def get_children_under_parent_id(self,
                                     parent_id,
                                     query_contains_string=None,
                                     query_is_string=None,
                                     max_results=None):

        _logger.info("Getting client for child-listing.")

        client = self.__auth.get_client()

        assert \
            (query_contains_string is not None and \
             query_is_string is not None) is False, \
            "The query_contains_string and query_is_string parameters are "\
            "mutually exclusive."

        if query_is_string:
            query = ("title='%s'" % 
                     (gdrivefs.gdfs.fsutility.escape_filename_for_query(query_is_string)))
        elif query_contains_string:
            query = ("title contains '%s'" % 
                     (gdrivefs.gdfs.fsutility.escape_filename_for_query(query_contains_string)))
        else:
            query = None

        _logger.info("Listing entries under parent with ID [%s].  QUERY= "
                     "[%s]", parent_id, query)

        response = client.children().list(
                    q=query, 
                    folderId=parent_id,
                    maxResults=max_results).execute()

        self.__assert_response_kind(response, 'drive#childList')

        return [ entry[u'id'] for entry in response[u'items'] ]

    @_marshall
    def get_entries(self, entry_ids):
        retrieved = { }
        for entry_id in entry_ids:
            retrieved[entry_id] = self.get_entry(entry_id)

        _logger.debug("(%d) entries were retrieved.", len(retrieved))

        return retrieved

    @_marshall
    def get_entry(self, entry_id):
        client = self.__auth.get_client()

        response = client.files().get(fileId=entry_id).execute()
        self.__assert_response_kind(response, 'drive#file')

        return \
            gdrivefs.gdtool.normal_entry.NormalEntry('direct_read', response)

    @_marshall
    def list_files(self, query_contains_string=None, query_is_string=None, 
                   parent_id=None):
        
        _logger.info("Listing all files. CONTAINS=[%s] IS=[%s] "
                     "PARENT_ID=[%s]",
                     query_contains_string 
                        if query_contains_string is not None 
                        else '<none>', 
                     query_is_string 
                        if query_is_string is not None 
                        else '<none>', 
                     parent_id 
                        if parent_id is not None 
                        else '<none>')

        client = self.__auth.get_client()

        query_components = []

        if parent_id:
            query_components.append("'%s' in parents" % (parent_id))

        if query_is_string:
            query_components.append("title='%s'" % 
                                    (gdrivefs.gdfs.fsutility.escape_filename_for_query(query_is_string)))
        elif query_contains_string:
            query_components.append("title contains '%s'" % 
                                    (gdrivefs.gdfs.fsutility.escape_filename_for_query(query_contains_string)))

        # Make sure that we don't get any entries that we would have to ignore.

        hidden_flags = gdrivefs.conf.Conf.get('hidden_flags_list_remote')
        if hidden_flags:
            for hidden_flag in hidden_flags:
                query_components.append("%s = false" % (hidden_flag))

        query = ' and '.join(query_components) if query_components else None

        page_token = None
        page_num = 0
        entries = []
        while 1:
            _logger.debug("Doing request for listing of files with page-"
                          "token [%s] and page-number (%d): %s",
                          page_token, page_num, query)

            result = client.files().list(q=query, pageToken=page_token).\
                        execute()

            self.__assert_response_kind(result, 'drive#fileList')

            _logger.debug("(%d) entries were presented for page-number "
                          "(%d).", len(result[u'items']), page_num)

            for entry_raw in result[u'items']:
                entry = \
                    gdrivefs.gdtool.normal_entry.NormalEntry(
                        'list_files', 
                        entry_raw)

                entries.append(entry)

            if u'nextPageToken' not in result:
                _logger.debug("No more pages in file listing.")
                break

            _logger.debug("Next page-token in file-listing is [%s].", 
                          result[u'nextPageToken'])

            page_token = result[u'nextPageToken']
            page_num += 1

        return entries

    @_marshall
    def download_to_local(self, output_file_path, normalized_entry, mime_type, 
                          allow_cache=True):
        """Download the given file. If we've cached a previous download and the 
        mtime hasn't changed, re-use. The third item returned reflects whether 
        the data has changed since any prior attempts.
        """

        _logger.info("Downloading entry with ID [%s] and mime-type [%s] to "
                     "[%s].", normalized_entry.id, mime_type, output_file_path)

        if mime_type != normalized_entry.mime_type and \
                mime_type not in normalized_entry.download_links:
            message = ("Entry with ID [%s] can not be exported to type [%s]. "
                       "The available types are: %s" % 
                       (normalized_entry.id, mime_type, 
                        ', '.join(normalized_entry.download_links.keys())))

            _logger.warning(message)
            raise gdrivefs.errors.ExportFormatError(message)

        gd_mtime_epoch = time.mktime(
                            normalized_entry.modified_date.timetuple())

        _logger.info("File will be downloaded to [%s].", output_file_path)

        use_cache = False
        if allow_cache and os.path.isfile(output_file_path):
            # Determine if a local copy already exists that we can use.
            stat_info = os.stat(output_file_path)

            if gd_mtime_epoch == stat_info.st_mtime:
                use_cache = True

        if use_cache:
            # Use the cache. It's fine.

            _logger.info("File retrieved from the previously downloaded, "
                         "still-current file.")

            return (stat_info.st_size, False)

        # Go and get the file.

        authed_http = self.__auth.get_authed_http()

        url = normalized_entry.download_links[mime_type]

        with open(output_file_path, 'wb') as f:
            downloader = gdrivefs.gdtool.chunked_download.ChunkedDownload(
                            f, 
                            authed_http, 
                            url)

            progresses = []

            while 1:
                status, done, total_size = downloader.next_chunk()
                assert status.total_size is not None, \
                       "total_size is None"

                _logger.debug("Read chunk: STATUS=[%s] DONE=[%s] "
                              "TOTAL_SIZE=[%s]", status, done, total_size)

                p = status.progress()

                _logger.debug("Chunk: PROGRESS=[%s] TOTAL-SIZE=[%s] "
                              "RESUMABLE-PROGRESS=[%s]",
                              p, status.total_size, status.resumable_progress)

                percent = p if status.total_size > 0 else 100.0
# TODO(dustin): This just places an arbitrary limit on the number of empty 
#               chunks we can receive. Can we drop this to 1?
                if len(progresses) >= _MAX_EMPTY_CHUNKS:
                    assert percent > progresses[0], \
                           "Too many empty chunks have been received."

                progresses.append(percent)

                # Constrain how many percentages we keep.
                if len(progresses) > _MAX_EMPTY_CHUNKS:
                    del progresses[0]

                if done is True:
                    break

            _logger.debug("Download complete. Offset is: (%d)", f.tell())

        os.utime(output_file_path, (time.time(), gd_mtime_epoch))

        return (total_size, True)

    @_marshall
    def create_directory(self, filename, parents, **kwargs):

        mimetype_directory = gdrivefs.conf.Conf.get('directory_mimetype')
        return self.__insert_entry(
                False,
                filename, 
                parents,
                mimetype_directory, 
                **kwargs)

    @_marshall
    def create_file(self, filename, parents, mime_type, data_filepath=None, 
                    **kwargs):
# TODO: It doesn't seem as if the created file is being registered.
        # Even though we're supposed to provide an extension, we can get away 
        # without having one. We don't want to impose this when acting like a 
        # normal FS.

        # If no data and no mime-type was given, default it.
        if mime_type == None:
            mime_type = gdrivefs.conf.Conf.get('file_default_mime_type')
            _logger.debug("No mime-type was presented for file "
                          "create/update. Defaulting to [%s].",
                          mime_type)

        return self.__insert_entry(
                True,
                filename,
                parents,
                mime_type,
                data_filepath=data_filepath,
                **kwargs)

    @_marshall
    def __insert_entry(self, is_file, filename, parents, mime_type, 
                       data_filepath=None, modified_datetime=None, 
                       accessed_datetime=None, is_hidden=False, 
                       description=None):

        if parents is None:
            parents = []

        now_phrase = gdrivefs.time_support.get_flat_normal_fs_time_from_dt()

        if modified_datetime is None:
            modified_datetime = now_phrase 
    
        if accessed_datetime is None:
            accessed_datetime = now_phrase 

        _logger.info("Creating entry with filename [%s] under parent(s) "
                     "[%s] with mime-type [%s]. MTIME=[%s] ATIME=[%s] "
                     "DATA_FILEPATH=[%s]",
                     filename, ', '.join(parents), mime_type, 
                     modified_datetime, accessed_datetime, data_filepath)

        client = self.__auth.get_client()

        ## Create request-body.

        body = { 
                'title': filename, 
                'parents': [dict(id=parent) for parent in parents], 
                'labels': { "hidden": is_hidden }, 
                'mimeType': mime_type,
            }

        if description is not None:
            body['description'] = description

        if modified_datetime is not None:
            body['modifiedDate'] = modified_datetime

        if accessed_datetime is not None:
            body['lastViewedByMeDate'] = accessed_datetime

        ## Create request-arguments.

        args = {
            'body': body,
        }

        if data_filepath:
            args.update({
                'media_body': 
                    apiclient.http.MediaFileUpload(
                        data_filepath, 
                        mimetype=mime_type, 
                        resumable=True,
                        chunksize=_DEFAULT_UPLOAD_CHUNK_SIZE_B),
# TODO(dustin): Documented, but does not exist.
#                'uploadType': 'resumable',
            })

        if gdrivefs.config.IS_DEBUG is True:
            _logger.debug("Doing file-insert with:\n%s", 
                          pprint.pformat(args))

        request = client.files().insert(**args)

        response = self.__finish_upload(
                    filename,
                    request,
                    data_filepath is not None)

        self.__assert_response_kind(response, 'drive#file')

        normalized_entry = \
            gdrivefs.gdtool.normal_entry.NormalEntry(
                'insert_entry', 
                response)

        _logger.info("New entry created with ID [%s].", normalized_entry.id)

        return normalized_entry

    @_marshall
    def truncate(self, normalized_entry):

        _logger.info("Truncating entry [%s].", normalized_entry.id)

        client = self.__auth.get_client()

        file_ = \
            apiclient.http.MediaFileUpload(
                '/dev/null',
                mimetype=normalized_entry.mime_type)

        args = { 
            'fileId': normalized_entry.id, 
# TODO(dustin): Can we omit 'body'?
            'body': {}, 
            'media_body': file_,
        }

        response = client.files().update(**args).execute()
        self.__assert_response_kind(response, 'drive#file')

        _logger.debug("Truncate complete: [%s]", normalized_entry.id)

        return response

    @_marshall
    def update_entry(self, normalized_entry, filename=None, data_filepath=None, 
                     mime_type=None, parents=None, modified_datetime=None, 
                     accessed_datetime=None, is_hidden=False, 
                     description=None):

        _logger.info("Updating entry [%s].", normalized_entry)

        client = self.__auth.get_client()

        # Build request-body.
        
        body = {}

        if mime_type is not None:
            body['mimeType'] = mime_type 
        else:
            body['mimeType'] = normalized_entry.mime_type

        if filename is not None:
            body['title'] = filename
        
        if parents is not None:
            body['parents'] = parents

        if is_hidden is not None:
            body['labels'] = { "hidden": is_hidden }

        if description is not None:
            body['description'] = description

        set_mtime = True
        if modified_datetime is not None:
            body['modifiedDate'] = modified_datetime
        else:
            body['modifiedDate'] = \
                gdrivefs.time_support.get_flat_normal_fs_time_from_dt()

        if accessed_datetime is not None:
            set_atime = True
            body['lastViewedByMeDate'] = accessed_datetime
        else:
            set_atime = False

        # Build request-arguments.

        args = { 
            'fileId': normalized_entry.id, 
            'body': body, 
            'setModifiedDate': set_mtime, 
            'updateViewedDate': set_atime,
        }

        if data_filepath is not None:
            _logger.debug("We'll be sending a file in the update: [%s] [%s]", 
                          normalized_entry.id, data_filepath)

            # We can only upload large files using resumable-uploads.
            args.update({
                'media_body': 
                    apiclient.http.MediaFileUpload(
                        data_filepath, 
                        mimetype=mime_type, 
                        resumable=True,
                        chunksize=_DEFAULT_UPLOAD_CHUNK_SIZE_B),
# TODO(dustin): Documented, but does not exist.
#                'uploadType': 'resumable',
            })

        _logger.debug("Sending entry update: [%s]", normalized_entry.id)

        request = client.files().update(**args)

        result = self.__finish_upload(
                    normalized_entry.title,
                    request,
                    data_filepath is not None)

        normalized_entry = \
            gdrivefs.gdtool.normal_entry.NormalEntry('update_entry', result)

        _logger.debug("Entry updated: [%s]", normalized_entry)

        return normalized_entry

    def __finish_upload(self, filename, request, has_file):
        """Finish a resumable-upload is a file was given, or just execute the 
        request if not.
        """

        if has_file is False:
            return request.execute()

        _logger.debug("We need to finish updating the entry's data: [%s]", 
                      filename)

        result = None
        while result is None:
            status, result = request.next_chunk()

            if status:
                if status.total_size == 0:
                    _logger.debug("Uploaded (zero-length): [%s]", filename)
                else:
                    _logger.debug("Uploaded [%s]: %.2f%%", 
                                  filename, status.progress() * 100)

        return result

    @_marshall
    def rename(self, normalized_entry, new_filename):

        result = gdrivefs.gdfs.fsutility.split_path_nolookups(new_filename)
        (path, filename_stripped, mime_type, is_hidden) = result

        _logger.debug("Renaming entry [%s] to [%s]. IS_HIDDEN=[%s]",
                      normalized_entry, filename_stripped, is_hidden)

        return self.update_entry(normalized_entry, filename=filename_stripped, 
                                 is_hidden=is_hidden)

    @_marshall
    def remove_entry(self, normalized_entry):

        _logger.info("Removing entry with ID [%s].", normalized_entry.id)

        client = self.__auth.get_client()

        args = { 'fileId': normalized_entry.id }

        try:
            result = client.files().delete(**args).execute()
        except Exception as e:
            if e.__class__.__name__ == 'HttpError' and \
               str(e).find('File not found') != -1:
                raise NameError(normalized_entry.id)

            _logger.exception("Could not send delete for entry with ID [%s].",
                              normalized_entry.id)
            raise

        _logger.info("Entry deleted successfully.")

_THREAD_STORAGE = None
def get_gdrive():
    """Return an instance of _GdriveManager unique to each thread (we can't 
    reuse sockets between threads).
    """

    global _THREAD_STORAGE

    if _THREAD_STORAGE is None:
        _THREAD_STORAGE = threading.local()

    try:
        return _THREAD_STORAGE.gm
    except AttributeError:
        _THREAD_STORAGE.gm = _GdriveManager()
        return _THREAD_STORAGE.gm



class GdFsError(Exception):
    pass


class AuthorizationError(GdFsError):
    """All authorization-related errors inherit from this."""
    pass


class AuthorizationFailureError(AuthorizationError):
    """There was a general authorization failure."""
    pass
        

class AuthorizationFaultError(AuthorizationError):
    """Our authorization is not available or has expired."""
    pass


class MustIgnoreFileError(GdFsError):
    """An error requiring us to ignore the file."""
    pass


class FilenameQuantityError(MustIgnoreFileError):
    """Too many filenames share the same name in a single directory."""
    pass


class ExportFormatError(GdFsError):
    """A format was not available for export."""
    pass


class GdNotFoundError(GdFsError):
    """A file/path was not found."""
    pass



import logging
import re
import dateutil.parser
import json
import time
import pprint

from time import mktime
from mimetypes import guess_type
from numbers import Number
from datetime import datetime

from gdrivefs.conf import Conf
from gdrivefs.utility import utility
from gdrivefs.errors import ExportFormatError
from gdrivefs.time_support import get_flat_normal_fs_time_from_dt

_logger = logging.getLogger(__name__)


class NormalEntry(object):
    __default_general_mime_type = Conf.get('default_mimetype')
    __directory_mimetype = Conf.get('directory_mimetype')

    __properties_extra = [
        'is_directory', 
        'is_visible', 
        'parents', 
        'download_types',
        'modified_date',
        'modified_date_epoch',
        'mtime_byme_date',
        'mtime_byme_date_epoch',
        'atime_byme_date',
        'atime_byme_date_epoch',
    ]

    def __init__(self, gd_resource_type, raw_data):
        self.__info = {}
        self.__parents = []
        self.__raw_data = raw_data
        self.__cache_data = None
        self.__cache_mimetypes = None
        self.__cache_dict = {}

        # Return True if reading from this file should return info and deposit 
        # the data elsewhere. This is predominantly determined by whether we 
        # can get a file-size up-front, or we have to decide on a specific 
        # mime-type in order to do so.

        requires_mimetype = u'fileSize' not in self.__raw_data and \
                            raw_data[u'mimeType'] != self.__directory_mimetype

        self.__info['requires_mimetype'] = \
            requires_mimetype
        
        self.__info['title'] = \
            raw_data[u'title']
        
        self.__info['mime_type'] = \
            raw_data[u'mimeType']
        
        self.__info['labels'] = \
            raw_data[u'labels']
        
        self.__info['id'] = \
            raw_data[u'id']
        
        self.__info['last_modifying_user_name'] = \
            raw_data[u'lastModifyingUserName']
        
        self.__info['writers_can_share'] = \
            raw_data[u'writersCanShare']

        self.__info['owner_names'] = \
            raw_data[u'ownerNames']
        
        self.__info['editable'] = \
            raw_data[u'editable']
        
        self.__info['user_permission'] = \
            raw_data[u'userPermission']

        self.__info['link'] = \
            raw_data.get(u'embedLink')
        
        self.__info['file_size'] = \
            int(raw_data.get(u'fileSize', 0))
        
        self.__info['file_extension'] = \
            raw_data.get(u'fileExtension')
        
        self.__info['md5_checksum'] = \
            raw_data.get(u'md5Checksum')
        
        self.__info['image_media_metadata'] = \
            raw_data.get(u'imageMediaMetadata')

        self.__info['download_links'] = \
            raw_data.get(u'exportLinks', {})

        try:
            self.__info['download_links'][self.__info['mime_type']] = \
                raw_data[u'downloadUrl']
        except KeyError:
            pass

        self.__update_display_name()

        for parent in raw_data[u'parents']:
            self.__parents.append(parent[u'id'])

    def __getattr__(self, key):
        return self.__info[key]

    def __str__(self):
        return ("<NORMAL ID= [%s] MIME= [%s] NAME= [%s] URIS= (%d)>" % 
                (self.id, self.mime_type, self.title, 
                 len(self.download_links)))

    def __repr__(self):
        return str(self)

    def __update_display_name(self):
        # This is encoded for displaying locally.
        self.__info['title_fs'] = utility.translate_filename_charset(self.__info['title'])

    def temp_rename(self, new_filename):
        """Set the name to something else, here, while we, most likely, wait 
        for the change at the server to propogate.
        """
    
        self.__info['title'] = new_filename
        self.__update_display_name()

    def normalize_download_mimetype(self, specific_mimetype=None):
        """If a mimetype is given, return it if there is a download-URL 
        available for it, or fail. Else, determine if a copy can downloaded 
        with the default mime-type (application/octet-stream, or something 
        similar), or return the only mime-type in the event that there's only 
        one download format.
        """

        if self.__cache_mimetypes is None:
            self.__cache_mimetypes = [[], None]
        
        if specific_mimetype is not None:
            if specific_mimetype not in self.__cache_mimetypes[0]:
                _logger.debug("Normalizing mime-type [%s] for download.  "
                              "Options: %s", 
                              specific_mimetype, self.download_types)

                if specific_mimetype not in self.download_links:
                    raise ExportFormatError("Mime-type [%s] is not available for "
                                            "download. Options: %s" % 
                                            (self.download_types))

                self.__cache_mimetypes[0].append(specific_mimetype)

            return specific_mimetype

        if self.__cache_mimetypes[1] is None:
            # Try to derive a mimetype from the filename, and see if it matches
            # against available export types.
            (mimetype_candidate, _) = guess_type(self.title_fs, True)
            if mimetype_candidate is not None and \
               mimetype_candidate in self.download_links:
                mime_type = mimetype_candidate

            elif NormalEntry.__default_general_mime_type in self.download_links:
                mime_type = NormalEntry.__default_general_mime_type

            # If there's only one download link, resort to using it (perhaps it was 
            # an uploaded file, assigned only one type).
            elif len(self.download_links) == 1:
                mime_type = self.download_links.keys()[0]

            else:
                raise ExportFormatError("A correct mime-type needs to be "
                                        "specified. Options: %s" % 
                                        (self.download_types))

            self.__cache_mimetypes[1] = mime_type

        return self.__cache_mimetypes[1]

    def __convert(self, data):
        if isinstance(data, dict):
            list_ = [("K(%s)=V(%s)" % (self.__convert(key), 
                                  self.__convert(value))) \
                     for key, value \
                     in data.iteritems()]

            final = '; '.join(list_)
            return final
        elif isinstance(data, list):
            final = ', '.join([('LI(%s)' % (self.__convert(element))) \
                               for element \
                               in data])
            return final
        elif isinstance(data, unicode):
            return utility.translate_filename_charset(data)
        elif isinstance(data, Number):
            return str(data)
        elif isinstance(data, datetime):
            return get_flat_normal_fs_time_from_dt(data)
        else:
            return data

    def get_data(self):
            original = dict([(key.encode('ASCII'), value) 
                                for key, value 
                                in self.__raw_data.iteritems()])

            distilled = self.__info

            extra = dict([(key, getattr(self, key)) 
                                for key 
                                in self.__properties_extra])

            data_dict = {'original': original,
                         #'distilled': distilled,
                         'extra': extra}

            return data_dict

    @property
    def xattr_data(self):
        if self.__cache_data is None:
            data_dict = self.get_data()
            
            attrs = {}
            for a_type, a_dict in data_dict.iteritems():
#                self.__log.debug("Setting [%s]." % (a_type))
                for key, value in a_dict.iteritems():
                    fqkey = ('user.%s.%s' % (a_type, key))
                    attrs[fqkey] = self.__convert(value)
 
            self.__cache_data = attrs

        return self.__cache_data

    @property
    def is_directory(self):
        """Return True if we represent a directory."""
        return (self.__info['mime_type'] == self.__directory_mimetype)

    @property
    def is_visible(self):
        if [ flag 
             for flag, value 
             in self.labels.items() 
             if flag in Conf.get('hidden_flags_list_local') and value ]:
            return False
        else:
            return True

    @property
    def parents(self):
        return self.__parents

    @property
    def download_types(self):
        return self.download_links.keys()

    @property
    def modified_date(self):
        if 'modified_date' not in self.__cache_dict:
            self.__cache_dict['modified_date'] = \
                dateutil.parser.parse(self.__raw_data[u'modifiedDate'])

        return self.__cache_dict['modified_date']

    @property
    def modified_date_epoch(self):
        # mktime() only works in terms of the local timezone, so compensate 
        # (this works with DST, too).
        return mktime(self.modified_date.timetuple()) - time.timezone
        
    @property  
    def mtime_byme_date(self):
        if 'modified_byme_date' not in self.__cache_dict:
            self.__cache_dict['modified_byme_date'] = \
                dateutil.parser.parse(self.__raw_data[u'modifiedByMeDate'])

        return self.__cache_dict['modified_byme_date']

    @property
    def mtime_byme_date_epoch(self):
        return mktime(self.mtime_byme_date.timetuple()) - time.timezone

    @property
    def atime_byme_date(self):
        if 'viewed_byme_date' not in self.__cache_dict:
            self.__cache_dict['viewed_byme_date'] = \
                dateutil.parser.parse(self.__raw_data[u'lastViewedByMeDate']) \
                if u'lastViewedByMeDate' in self.__raw_data \
                else None

        return self.__cache_dict['viewed_byme_date']

    @property
    def atime_byme_date_epoch(self):
        return mktime(self.atime_byme_date.timetuple()) - time.timezone \
                if self.atime_byme_date \
                else None



from unittest import TestCase, main

from gdrivefs.gdtool import Conf

class ConfTestCase(TestCase):
    """Test the Conf class."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_config(self):
        """Test for the existence of all configuration keys."""

        keys = [ 'auth_temp_path',
                 'auth_cache_filename',
                 'auth_secrets_filepath',
                 'change_check_interval_s']
        try:
            for key in keys:
                Conf.get(key)
        except (Exception) as e:
            self.fail("Could not retrieve value for configuration key [%s]." % 
                      (key))

if __name__ == '__main__':
    main()




from unittest import TestCase, main

from gdrivefs.gdtool import drive_proxy, AccountInfo
from gdrivefs.cache import EntryCache, PathRelations
from gdrivefs.utility import get_utility

class GetDriveTestCase(TestCase):
    """Test the _GdriveManager class via _GoogleProxy via get_drive()."""

    drive_proxy = None

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_list_files(self):
        """Test the list_files() call on the Drive object."""

#        files = drive_proxy('list_files')

#        print(files)

    def test_list_files_by_parent_id(self):

        return

        entries = drive_proxy('list_files')

        from pprint import pprint
        import json
        with open('/tmp/entries', 'w') as f:
            for entry in entries:
                f.write("%s\n" % (json.dumps(entry.info)))

    def test_get_changes(self):

        from gdrivefs.change import get_change_manager
        get_change_manager().process_updates()

        import sys
        sys.exit()

        (largest_change_id, next_page_token, changes) = drive_proxy('list_changes')

        print("Largest Change ID: [%s]" % (largest_change_id))
        print("Next Page Token: [%s]" % (next_page_token))

        from pprint import pprint
        pprint(len(changes))
        print

        for change_id, (entry_id, was_deleted, entry) in changes.iteritems():
            print("%d> [%s] D:[%s] [%s]" % (change_id, entry_id, was_deleted, entry.title if entry else '<deleted>'))

#            pprint(change_id)
#            pprint(change_info)

#            import sys
#            sys.exit()
#            pprint("%s, %s, %s" % (type(change[0]), type(change[1]), type(change[2])))

    def test_get_parents_containing_id(self):

        return
        
        entry_id = u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk'

        try:
            parent_ids = drive_proxy('get_parents_containing_id', 
                                     child_id=entry_id)
        except:
            logging.exception("Could not retrieve parents for child with ID "
                              "[%s]." % (entry_id))
            raise

        from pprint import pprint
        pprint(parent_ids)

    def test_download_file(self):

        return

        from gdrivefs.gdtool import drive_proxy
        http = drive_proxy('get_authed_http')

        normalized_entry = EntryCache.get_instance().cache.get('1DcIWAjj-pnSCXBQa3kHJQuL-QMRoopx8Yx_LVhfRigk')
        mime_type = 'text/plain'

        files = drive_proxy('download_to_local', normalized_entry=normalized_entry, mime_type=mime_type)

        return

        from pprint import pprint
        url = files[16].download_links[u'text/plain']
        pprint(url)

        data = http.request(url)
        response_headers = data[0]

        import re
        r = re.compile('Range')
        found = [("%s: %s" % (k, v)) for k, v in response_headers.iteritems() if r.match(k)]
        if found:
            print("Found: %s" % (", ".join(found)))

        print(">>>===============================================")
#        print(data[1][:200])
        print("<<<===============================================")

    def test_get_about(self):

        return

        entry_id_1 = u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk'
        entry1 = EntryCache.get_instance().cache.get(entry_id_1)
#        result = PathRelations.get_instance().register_entry(entry1)

        entry_id_2 = u'0AJFt2OXeDBqSUk9PVA'
#        entry2 = EntryCache.get_instance().cache.get(entry_id_2)
#        result = PathRelations.get_instance().register_entry(entry2)

        path_relations = PathRelations.get_instance()

        #print(len(entry.parents))
#        path_relations.dump_ll()

#        print(AccountInfo().root_id)

#        path_relations.dump_entry_clause('0AJFt2OXeDBqSUk9PVA')
#        PathRelations.get_instance().dump_entry_clause('0B5Ft2OXeDBqSSmdIek1aajZtVDA')
#        return
#        entry_clause = path_relations.get_entry_clause_by_id(entry_id_1)
        #result = path_relations.find_path_components_goandget('/')

        result = path_relations.get_child_filenames_from_entry_id(entry_id_2)

        from pprint import pprint
        pprint(result)

#        result = EntryCache.get_instance().cache.get(u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk')
#        result = EntryCache.get_instance().cache.get(u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk')
#        result = EntryCache.get_instance().cache.get(u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk')
#        result = EntryCache.get_instance().cache.get(u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk')
#        print(result)
        return
        

#        result = AccountInfo().root_id

        #about = drive_proxy('get_about_info')

#        entries = drive_proxy('get_children_under_parent_id', parent_id=about.root_id)
        #entries = drive_proxy('get_parents_over_child_id', child_id=u'11EIs1ZxCykme0FnAdY8Xm_ktUCQ9y5lHC3EwAKFsiFk')


#        print(response[u'rootFolderId'])
        import pprint
#        pprint.pprint(response[u'importFormats'])
        pprint.pprint(result)

    def test_remove_entry(self):

        return

        from gdrivefs.cache import PathRelations

        path_relations = PathRelations.get_instance()
        entry_clause = path_relations.get_clause_from_path('HelloFax')

        filenames = path_relations.get_child_filenames_from_entry_id(entry_clause[3])
        
        root_id = u'0AJFt2OXeDBqSUk9PVA'
        middle_id = entry_clause[3]
        child_id = u'0B5Ft2OXeDBqSTmpjSHlVbEV5ajg'

#        from pprint import pprint
#        pprint(filenames)

#        path_relations.dump_entry_clause(middle_id)

        print("1: =============================")
        path_relations.dump_ll()
        print("2: =============================")
#        print("middle: %s" % (middle_id))
#        return

        path_relations.remove_entry_recursive(middle_id)
#        path_relations.remove_entry(middle_id)

        print("3: =============================")
        path_relations.dump_ll()

        return

        try:
            path_relations.dump_entry_clause(root_id)
        except:
            print("<No root.>")

        try:
            path_relations.dump_entry_clause(middle_id)
        except:
            print("<No middle.>")

        try:
            path_relations.dump_entry_clause(child_id)
        except:
            print("<No child.>")

    def test_insert_entry(self):

        import datetime
#        filename = ("NewFolder_%s" % (datetime.datetime.now().strftime("%H%M%S")))
#        entry = drive_proxy('create_directory', filename=filename)

        filename = ("NewFile_%s.txt" % (datetime.datetime.now().strftime("%H%M%S")))
        entry = drive_proxy('create_file', filename=filename, data_filepath='/tmp/tmpdata.txt', parents=[])

        print(entry.id)

if __name__ == '__main__':
    main()





import logging
import threading
import time
import datetime

#import gdrivefs.report

import gdrivefs.state

from gdrivefs.conf import Conf
from gdrivefs.cache.cache_registry import CacheRegistry, CacheFault

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)


class CacheAgent(object):
    """A particular namespace within the cache."""

    registry        = None
    resource_name   = None
    max_age         = None

    fault_handler       = None
    cleanup_pretrigger  = None

    report              = None
    report_source_name  = None

    def __init__(self, resource_name, max_age, fault_handler=None, 
                 cleanup_pretrigger=None):
        _logger.debug("CacheAgent(%s,%s,%s,%s)" % (resource_name, max_age, 
                                                   type(fault_handler), 
                                                   cleanup_pretrigger))

        self.registry = CacheRegistry.get_instance(resource_name)
        self.resource_name = resource_name
        self.max_age = max_age

        self.fault_handler = fault_handler
        self.cleanup_pretrigger = cleanup_pretrigger

#        self.report = Report.get_instance()
#        self.report_source_name = ("cache-%s" % (self.resource_name))

        self.__t = None
        self.__t_quit_ev = threading.Event()

        self.__start_cleanup()

    def __del__(self):
        self.__stop_cleanup()

# TODO(dustin): Currently disabled. The system doesn't rely on it, and it's 
#               just another thread that unnecessarily runs, and trips up our 
#               ability to test individual components in simple isolation. It
#               needs to be refactored.
#
#               We'd like to either refactor into a multiprocessing worker, or
#               just send to statsd (which would be kindof cool).
#        self.__post_status()

#    def __del__(self):
#
#        if self.report.is_source(self.report_source_name):
#            self.report.remove_all_values(self.report_source_name)
#        pass

#    def __post_status(self):
#        """Send the current status to our reporting tool."""
#
#        num_values = self.registry.count(self.resource_name)
#
#        self.report.set_values(self.report_source_name, 'count', 
#                               num_values)
#
#        status_post_interval_s = Conf.get('cache_status_post_frequency_s')
#        status_timer = Timer(status_post_interval_s, self.__post_status)
#
#        Timers.get_instance().register_timer('status', status_timer)

    def __cleanup(self):
        """Scan the current cache and determine items old-enough to be 
        removed.
        """

        cleanup_interval_s = Conf.get('cache_cleanup_check_frequency_s')

        _logger.info("Cache-cleanup thread running: %s", self)

        while self.__t_quit_ev.is_set() is False and \
                  gdrivefs.state.GLOBAL_EXIT_EVENT.is_set() is False:
            _logger.debug("Doing clean-up for cache resource with name [%s]." % 
                          (self.resource_name))

            cache_dict = self.registry.list_raw(self.resource_name)

            total_keys = [ (key, value_tuple[1]) for key, value_tuple \
                                in cache_dict.iteritems() ]

            cleanup_keys = [ key for key, value_tuple \
                                in cache_dict.iteritems() \
                                if (datetime.datetime.now() - value_tuple[1]).seconds > \
                                        self.max_age ]

            _logger.debug("Found (%d) entries to clean-up from entry-cache." % 
                          (len(cleanup_keys)))

            if cleanup_keys:
                for key in cleanup_keys:
                    _logger.debug("Cache entry [%s] under resource-name [%s] "
                                  "will be cleaned-up." % 
                                  (key, self.resource_name))

                    if self.exists(key, no_fault_check=True) == False:
                        _logger.debug("Entry with ID [%s] has already been "
                                      "cleaned-up." % (key))
                    else:
                        self.remove(key)
            else:
                _logger.debug("No cache-cleanup required.")
                time.sleep(cleanup_interval_s)

        _logger.info("Cache-cleanup thread terminating: %s", self)

    def __start_cleanup(self):
        _logger.info("Starting cache-cleanup thread: %s", self)

        self.__t = threading.Thread(target=self.__cleanup)
        self.__t.start()

    def __stop_cleanup(self):
        _logger.info("Stopping cache-cleanup thread: %s", self)

        self.__t_quit_ev.set()
        self.__t.join()

    def set(self, key, value):
        _logger.debug("CacheAgent.set(%s,%s)" % (key, value))

        return self.registry.set(self.resource_name, key, value)

    def remove(self, key):
        _logger.debug("CacheAgent.remove(%s)" % (key))

        return self.registry.remove(self.resource_name, 
                                    key, 
                                    cleanup_pretrigger=self.cleanup_pretrigger)

    def get(self, key, handle_fault = None):

        if handle_fault == None:
            handle_fault = True

        _logger.debug("CacheAgent.get(%s)" % (key))

        try:
            result = self.registry.get(self.resource_name, 
                                       key, 
                                       max_age=self.max_age, 
                                       cleanup_pretrigger=self.cleanup_pretrigger)
        except CacheFault:
            _logger.debug("There was a cache-miss while requesting item with "
                          "ID (key).")

            if self.fault_handler == None or not handle_fault:
                raise

            result = self.fault_handler(self.resource_name, key)
            if result is None:
                raise

        return result

    def exists(self, key, no_fault_check=False):
        _logger.debug("CacheAgent.exists(%s)" % (key))

        return self.registry.exists(self.resource_name, key, 
                                    max_age=self.max_age,
                                    cleanup_pretrigger=self.cleanup_pretrigger,
                                    no_fault_check=no_fault_check)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __delitem__(self, key):
        return self.remove(key)



import logging
import pickle
import json

from oauth2client.client import flow_from_clientsecrets
from oauth2client.client import OOB_CALLBACK_URN

from datetime import datetime
from httplib2 import Http
from tempfile import NamedTemporaryFile
from os import remove

from gdrivefs.errors import AuthorizationFailureError, AuthorizationFaultError
from gdrivefs.conf import Conf

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)


class _OauthAuthorize(object):
    """Manages authorization process."""

    def __init__(self):
        cache_filepath  = Conf.get('auth_cache_filepath')
        api_credentials = Conf.get('api_credentials')

        self.cache_filepath = cache_filepath
        self.credentials = None

        with NamedTemporaryFile() as f:
            json.dump(api_credentials, f)
            f.flush()

            self.flow = flow_from_clientsecrets(f.name, 
                                                scope=self.__get_scopes(), 
                                                redirect_uri=OOB_CALLBACK_URN)
        
    def __get_scopes(self):
        scopes = "https://www.googleapis.com/auth/drive "\
                 "https://www.googleapis.com/auth/drive.file"
        return scopes

    def step1_get_auth_url(self):
        return self.flow.step1_get_authorize_url()

    def __clear_cache(self):
        if self.cache_filepath is not None:
            try:
                remove(self.cache_filepath)
            except:
                pass
    
    def __refresh_credentials(self):
        _logger.info("Doing credentials refresh.")

        http = Http()

        try:
            self.credentials.refresh(http)
        except:
            raise AuthorizationFailureError("Could not refresh credentials.")

        self.__update_cache(self.credentials)
            
        _logger.debug("Credentials have been refreshed.")
            
    def __step2_check_auth_cache(self):
        # Attempt to read cached credentials.

        if self.cache_filepath is None:
            raise ValueError("Credentials file-path is not set.")

        if self.credentials is None:
            _logger.debug("Checking for cached credentials: %s",
                          self.cache_filepath)

            with open(self.cache_filepath) as cache:
                credentials_serialized = cache.read()

            # If we're here, we have serialized credentials information.
            
            try:
                credentials = pickle.loads(credentials_serialized)
            except:
                # We couldn't decode the credentials. Kill the cache.
                self.__clear_cache()
                raise

            self.credentials = credentials
                
            # Credentials restored. Check expiration date.

            expiry_phrase = self.credentials.token_expiry.strftime(
                                '%Y%m%d-%H%M%S')
                
            _logger.debug("Cached credentials found with expire-date [%s].",
                          expiry_phrase)
            
            self.check_credential_state()

        return self.credentials

    def check_credential_state(self):
        """Do all of the regular checks necessary to keep our access going, 
        such as refreshing when we expire.
        """
        if(datetime.today() >= self.credentials.token_expiry):
            _logger.info("Credentials have expired. Attempting to refresh "
                         "them.")
            
            self.__refresh_credentials()
            return self.credentials

    def get_credentials(self):
        return self.__step2_check_auth_cache()
    
    def __update_cache(self, credentials):
        if self.cache_filepath is None:
            raise ValueError("Credentials file-path is not set.")

        # Serialize credentials.

        credentials_serialized = pickle.dumps(credentials)

        # Write cache file.

        with open(self.cache_filepath, 'w') as cache:
            cache.write(credentials_serialized)

    def step2_doexchange(self, auth_code):
        # Do exchange.

        _logger.debug("Doing exchange.")
        
        try:
            credentials = self.flow.step2_exchange(auth_code)
        except Exception as e:
            message = \
                "Could not do auth-exchange (this was either a legitimate "\
                "error, or the auth-exchange was attempted when not "\
                "necessary): {}".format(e)

            raise AuthorizationFailureError(message)
        
        _logger.debug("Credentials established.")

        self.__update_cache(credentials)
        self.credentials = credentials

oauth = None
def get_auth():
    global oauth
    if oauth is None:
        _logger.debug("Creating OauthAuthorize.")
        oauth = _OauthAuthorize()
    
    return oauth




import logging
import threading
import time

import gdrivefs.state

from gdrivefs.conf import Conf
from gdrivefs.gdtool.account_info import AccountInfo
from gdrivefs.gdtool.drive import get_gdrive
from gdrivefs.cache.volume import PathRelations, EntryCache

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.WARNING)


class _ChangeManager(object):
    def __init__(self):
        self.at_change_id = AccountInfo.get_instance().largest_change_id
        _logger.debug("Latest change-ID at startup is (%d)." % 
                      (self.at_change_id))

        self.__t = None
        self.__t_quit_ev = threading.Event()

    def mount_init(self):
        """Called when filesystem is first mounted."""

        self.__start_check()

    def mount_destroy(self):
        """Called when the filesystem is unmounted."""

        self.__stop_check()

    def __check_changes(self):
        _logger.info("Change-processing thread running.")

        interval_s = Conf.get('change_check_frequency_s')
        cm = get_change_manager()

        while self.__t_quit_ev.is_set() is False and \
                gdrivefs.state.GLOBAL_EXIT_EVENT.is_set() is False:
            _logger.debug("Checking for changes.")

            try:
                is_done = cm.process_updates()
            except:
                _logger.exception("Squelching an exception that occurred "
                                  "while reading/processing changes.")

                # Force another check, soon.
                is_done = False

            # If there are still more changes, take them as quickly as 
            # possible.
            if is_done is True:
                _logger.debug("No more changes. Waiting.")
                time.sleep(interval_s)
            else:
                _logger.debug("There are more changes to be applied. Cycling "
                              "immediately.")

        _logger.info("Change-processing thread terminating.")

    def __start_check(self):
        _logger.info("Starting change-processing thread.")

        self.__t = threading.Thread(target=self.__check_changes)
        self.__t.start()

    def __stop_check(self):
        _logger.info("Stopping change-processing thread.")

        self.__t_quit_ev.set()
        self.__t.join()

    def process_updates(self):
        """Process any changes to our files. Return True if everything is up to
        date or False if we need to be run again.
        """
# TODO(dustin): Reimplement using the "watch" interface. We'll have to find 
#               more documentation:
#
#               https://developers.google.com/drive/v2/reference/changes/watch
#
        start_at_id = (self.at_change_id + 1)

        gd = get_gdrive()
        result = gd.list_changes(start_change_id=start_at_id)

        (largest_change_id, next_page_token, changes) = result

        _logger.debug("The latest reported change-ID is (%d) and we're "
                      "currently at change-ID (%d).",
                      largest_change_id, self.at_change_id)

        _logger.info("(%d) changes will now be applied." % (len(changes)))

        for change_id, change_tuple in changes:
            # Apply the changes. We expect to be running them from oldest to 
            # newest.

            _logger.info("========== Change with ID (%d) will now be applied. ==========" %
                            (change_id))

            try:
                self.__apply_change(change_id, change_tuple)
            except:
                _logger.exception("There was a problem while processing change"
                                  " with ID (%d). No more changes will be "
                                  "applied." % (change_id))
                return False

            self.at_change_id = change_id

        return (next_page_token is None)

    def __apply_change(self, change_id, change_tuple):
        """Apply changes to our filesystem reported by GD. All we do is remove 
        the current record components, if it's valid, and then reload it with 
        what we were given. Note that since we don't necessarily know
        about the entries that have been changed, this also allows us to slowly
        increase our knowledge of the filesystem (of, obviously, only those 
        things that change).
        """

        (entry_id, was_deleted, entry) = change_tuple
        
        is_visible = entry.is_visible if entry else None

        _logger.info("Applying change with change-ID (%d), entry-ID [%s], "
                     "and is-visible of [%s]",
                     change_id, entry_id, is_visible)

        # First, remove any current knowledge from the system.

        _logger.debug("Removing all trace of entry with ID [%s] "
                      "(apply_change).", entry_id)

        PathRelations.get_instance().remove_entry_all(entry_id)

        # If it wasn't deleted, add it back.

        _logger.debug("Registering changed entry with ID [%s].", entry_id)

        if is_visible:
            path_relations = PathRelations.get_instance()
            path_relations.register_entry(entry)

_instance = None
def get_change_manager():
    global _instance

    if _instance is None:
        _instance = _ChangeManager()

    return _instance



import logging
from apiclient.discovery import DISCOVERY_URI

_logger = logging.getLogger(__name__)

# TODO(dustin): Move this module to the *config* directory, eliminate this 
#               class, and use the module directly.


class Conf(object):
    """Manages options."""

    api_credentials = {
        "web": { "client_id": "1056816309698.apps.googleusercontent.com",
                 "client_secret": "R7FJFlbtWXgUoG3ZjIAWUAzv",
                 "redirect_uris": [],
                 "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                 "token_uri": "https://accounts.google.com/o/oauth2/token"
               }}
    
    auth_cache_filepath                 = None
    gd_to_normal_mapping_filepath       = '/etc/gdfs/mime_mapping.json'
    extension_mapping_filepath          = '/etc/gdfs/extension_mapping.json'
    query_decay_intermed_prefix_length  = 7
    file_jobthread_max_idle_time        = 60
    file_chunk_size_kb                  = 1024
    file_download_temp_max_age_s        = 86400
    file_default_mime_type              = 'application/octet-stream'
    change_check_frequency_s            = 3
    hidden_flags_list_local             = [u'trashed', u'restricted']
    hidden_flags_list_remote            = [u'trashed']
    cache_cleanup_check_frequency_s     = 60
    cache_entries_max_age               = 8 * 60 * 60
    cache_status_post_frequency_s       = 10

# Deimplementing report functionality.
#    report_emit_frequency_s             = 60

    google_discovery_service_url        = DISCOVERY_URI
    default_buffer_read_blocksize       = 65536
    default_mimetype                    = 'application/octet-stream'
    directory_mimetype                  = u'application/vnd.google-apps.folder'
    default_perm_folder                 = '777'
    default_perm_file_editable          = '666'
    default_perm_file_noneditable       = '444'

    # How many extra entries to retrieve when an entry is accessed that is not
    # currently cached.
    max_readahead_entries = 10

    @staticmethod
    def get(key):
        return Conf.__dict__[key]

    @staticmethod
    def set(key, value):
        if key not in Conf.__dict__:
            raise KeyError(key)

        setattr(Conf, key, value)




import logging

from gdrivefs.cache.cache_agent import CacheAgent

_logger = logging.getLogger(__name__)


class CacheClientBase(object):
    """Meant to be inherited by a class. Is used to configure a particular 
    namespace within the cache.
    """



# TODO(dustin): This is a terrible object, and needs to be refactored. It 
#               doesn't provide any way to cleanup itself or CacheAgent, or any 
#               way to invoke a singleton of CacheAgent whose thread we can 
#               easier start or stop. Since this larger *wraps* CacheAgent, we 
#               might just dispose of it.



    @property
    def cache(self):
        try:
            return self.__cache
        except:
            pass

        self.__cache = CacheAgent(self.child_type, self.max_age, 
                                 fault_handler=self.fault_handler, 
                                 cleanup_pretrigger=self.cleanup_pretrigger)

        return self.__cache

    def __init__(self):
        child_type = self.__class__.__bases__[0].__name__
        max_age = self.get_max_cache_age_seconds()
        
        _logger.debug("CacheClientBase(%s,%s)" % (child_type, max_age))

        self.child_type = child_type
        self.max_age = max_age

        self.init()

    def __del__(self):
        del self.__cache

    def fault_handler(self, resource_name, key):
        pass

    def cleanup_pretrigger(self, resource_name, key, force):
        pass

    def init(self):
        pass

    def get_max_cache_age_seconds(self):
        raise NotImplementedError("get_max_cache_age() must be implemented in "
                                  "the CacheClientBase child.")

    @classmethod
    def get_instance(cls):
        """A helper method to dispense a singleton of whomever is inheriting "
        from us.
        """

        class_name = cls.__name__

        try:
            CacheClientBase.__instances
        except:
            CacheClientBase.__instances = { }

        try:
            return CacheClientBase.__instances[class_name]
        except:
            CacheClientBase.__instances[class_name] = cls()
            return CacheClientBase.__instances[class_name]





import logging
import time
import random

import oauth2client
import apiclient.http
import apiclient.errors

DEFAULT_CHUNK_SIZE = 1024 * 512

_logger = logging.getLogger(__name__)

# TODO(Dustin): Refactor this to be nice. It's largely just copy+pasted.


class ChunkedDownload(object):
    """"Download an entry, chunk by chunk. This code is mostly identical to
    MediaIoBaseDownload, which couldn't be used because we have a specific URL
    that needs to be downloaded (not a request object, which doesn't apply here).
    """

    @oauth2client.util.positional(4)
    def __init__(self, fd, http, uri, chunksize=DEFAULT_CHUNK_SIZE, start_at=0):
        """Constructor.

        Args:
          fd: io.Base or file object, The stream in which to write the downloaded
            bytes.
          http: The httplib2 resource.
          uri: The URL to be downloaded.
          chunksize: int, File will be downloaded in chunks of this many bytes.
        """

        self._fd = fd
        self._http = http
        self._uri = uri
        self._chunksize = chunksize
        self._progress = start_at
        self._total_size = None
        self._done = False

        # Stubs for testing.
        self._sleep = time.sleep
        self._rand = random.random

    @oauth2client.util.positional(1)
    def next_chunk(self, num_retries=0):
        """Get the next chunk of the download.

        Args:
          num_retries: Integer, number of times to retry 500's with randomized
                exponential backoff. If all retries fail, the raised HttpError
                represents the last request. If zero (default), we attempt the
                request only once.

        Returns:
          (status, done): (MediaDownloadStatus, boolean)
             The value of 'done' will be True when the media has been fully
             downloaded.

        Raises:
          apiclient.errors.HttpError if the response was not a 2xx.
          httplib2.HttpLib2Error if a transport error has occured.
        """

        headers = {
            'range': 'bytes=%d-%d' % (
                self._progress, self._progress + self._chunksize)
            }

        for retry_num in xrange(num_retries + 1):
            _logger.debug("Attempting to read chunk. ATTEMPT=(%d)/(%d)", 
                          retry_num + 1, num_retries + 1)

            if retry_num > 0:
                self._sleep(self._rand() * 2**retry_num)
                _logger.warning("Retry #%d for media download: GET %s, "
                                "following status: %d", 
                                retry_num, self._uri, resp.status)

            resp, content = self._http.request(self._uri, headers=headers)
            if resp.status < 500:
                break

        _logger.debug("Received chunk of size (%d).", len(content))

        if resp.status in [200, 206]:
            try:
                if resp['content-location'] != self._uri:
                    self._uri = resp['content-location']
            except KeyError:
                pass

            received_size_b = len(content)
            self._progress += received_size_b
            self._fd.write(content)

            # This seems to be the most correct method to get the filesize, but 
            # we've seen it not exist.
            if 'content-range' in resp:
                if self._total_size is None:
                    content_range = resp['content-range']
                    length = content_range.rsplit('/', 1)[1]
                    length = int(length)

                    self._total_size = length

                    _logger.debug("Received download size (content-range): "
                                  "(%d)", self._total_size)

            # There's a chance that "content-range" will be omitted for zero-
            # length files (or maybe files that are complete within the first 
            # chunk).

            else:
# TODO(dustin): Is this a valid assumption, or should it be an error?
                _logger.warning("No 'content-range' found in response. "
                                "Assuming that we've received all data.")

                self._total_size = received_size_b

# TODO(dustin): We were using this for a while, but it appears to be no larger 
#               then a single chunk.
#
#            # This method doesn't seem documented, but we've seen cases where 
#            # this is available, but "content-range" isn't.
#            if 'content-length' in resp:
#                self._total_size = int(resp['content-length'])
#
#                _logger.debug("Received download size (content-length): "
#                              "(%d)", self._total_size)


            assert self._total_size is not None, \
                   "File-size was not provided."

            _logger.debug("Checking if done. PROGRESS=(%d) TOTAL-SIZE=(%d)", 
                          self._progress, self._total_size)

            if self._progress == self._total_size:
                self._done = True

            return (apiclient.http.MediaDownloadProgress(
                        self._progress, 
                        self._total_size), \
                    self._done, \
                    self._total_size)
        else:
            raise apiclient.errors.HttpError(resp, content, uri=self._uri)



import logging
import json
import re
import sys

import gdrivefs.conf

_logger = logging.getLogger(__name__)

# TODO(dustin): Make these individual functions.


class _DriveUtility(object):
    """General utility functions loosely related to GD."""

    # Mime-types to translate to, if they appear within the "exportLinks" list.
    gd_to_normal_mime_mappings = {
            'application/vnd.google-apps.document':        
                'text/plain',
            'application/vnd.google-apps.spreadsheet':     
                'application/vnd.ms-excel',
            'application/vnd.google-apps.presentation':    
                'application/vnd.ms-powerpoint',
            'application/vnd.google-apps.drawing':         
                'application/pdf',
            'application/vnd.google-apps.audio':           
                'audio/mpeg',
            'application/vnd.google-apps.photo':           
                'image/png',
            'application/vnd.google-apps.video':           
                'video/x-flv'
        }

    # Default extensions for mime-types.
    default_extensions = { 
            'text/plain':                       'txt',
            'application/vnd.ms-excel':         'xls',
            'application/vnd.ms-powerpoint':    'ppt',
            'application/pdf':                  'pdf',
            'audio/mpeg':                       'mp3',
            'image/png':                        'png',
            'video/x-flv':                      'flv'
        }

    local_character_set = sys.getfilesystemencoding()

    def __init__(self):
        self.__load_mappings()

    def __load_mappings(self):
        # Allow someone to override our default mappings of the GD types.

        gd_to_normal_mapping_filepath = \
            gdrivefs.conf.Conf.get('gd_to_normal_mapping_filepath')

        try:
            with open(gd_to_normal_mapping_filepath, 'r') as f:
                self.gd_to_normal_mime_mappings.extend(json.load(f))
        except IOError:
            _logger.info("No mime-mapping was found.")

        # Allow someone to set file-extensions for mime-types, and not rely on 
        # Python's educated guesses.

        extension_mapping_filepath = \
            gdrivefs.conf.Conf.get('extension_mapping_filepath')

        try:
            with open(extension_mapping_filepath, 'r') as f:
                self.default_extensions.extend(json.load(f))
        except IOError:
            _logger.info("No extension-mapping was found.")

    def get_first_mime_type_by_extension(self, extension):

        found = [ mime_type 
                    for mime_type, temp_extension 
                    in self.default_extensions.iteritems()
                    if temp_extension == extension ]

        if not found:
            return None

        return found[0]

    def translate_filename_charset(self, original_filename):
        """Convert the given filename to the correct character set."""

        # fusepy doesn't support the Python 2.x Unicode type. Expect a native
        # string (anything but a byte string).
        return original_filename
       
#        # If we're in an older version of Python that still defines the Unicode
#        # class and the filename isn't unicode, translate it.
#
#        try:
#            sys.modules['__builtin__'].unicode
#        except AttributeError:
#            pass
#        else:
#            if issubclass(original_filename.__class__, unicode) is False:
#                return unicode(original_filename)#original_filename.decode(self.local_character_set)
#
#        # It's already unicode. Don't do anything.
#        return original_filename

    def make_safe_for_filename(self, text):
        """Remove any filename-invalid characters."""
    
        return re.sub('[^a-z0-9\-_\.]+', '', text)

utility = _DriveUtility()




import logging

from threading import RLock
from datetime import datetime

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)


class CacheFault(Exception):
    pass


class CacheRegistry(object):
    """The main cache container."""

    __rlock = RLock()

    def __init__(self):
        self.__cache = { }

    @staticmethod
    def get_instance(resource_name):
    
        with CacheRegistry.__rlock:
            try:
                CacheRegistry.__instance;
            except:
                CacheRegistry.__instance = CacheRegistry()

            if resource_name not in CacheRegistry.__instance.__cache:
                CacheRegistry.__instance.__cache[resource_name] = { }

        return CacheRegistry.__instance

    def set(self, resource_name, key, value):

        _logger.debug("CacheRegistry.set(%s,%s,%s)" % 
                      (resource_name, key, value))

        with CacheRegistry.__rlock:
            try:
                old_tuple = self.__cache[resource_name][key]
            except:
                old_tuple = None

            self.__cache[resource_name][key] = (value, datetime.now())

        return old_tuple

    def remove(self, resource_name, key, cleanup_pretrigger=None):

        _logger.debug("CacheRegistry.remove(%s,%s,%s)" % 
                      (resource_name, key, type(cleanup_pretrigger)))

        with CacheRegistry.__rlock:
            old_tuple = self.__cache[resource_name][key]

            self.__cleanup_entry(
                resource_name, 
                key, 
                True, 
                cleanup_pretrigger=cleanup_pretrigger)

        return old_tuple[0]

    def get(self, resource_name, key, max_age, cleanup_pretrigger=None):
        
        trigger_given_phrase = ('None' 
                                if cleanup_pretrigger == None 
                                else '<given>')

        _logger.debug("CacheRegistry.get(%s,%s,%s,%s)" % 
                      (resource_name, key, max_age, trigger_given_phrase))

        with CacheRegistry.__rlock:
            try:
                (value, timestamp) = self.__cache[resource_name][key]
            except:
                raise CacheFault("NonExist")

            if max_age != None and \
               (datetime.now() - timestamp).seconds > max_age:
                self.__cleanup_entry(resource_name, key, False, 
                                     cleanup_pretrigger=cleanup_pretrigger)
                raise CacheFault("Stale")

        return value

    def list_raw(self, resource_name):
        
        _logger.debug("CacheRegistry.list(%s)" % (resource_name))

        with CacheRegistry.__rlock:
            return self.__cache[resource_name]

    def exists(self, resource_name, key, max_age, cleanup_pretrigger=None, 
               no_fault_check=False):

        _logger.debug("CacheRegistry.exists(%s,%s,%s,%s)" % 
                      (resource_name, key, max_age, cleanup_pretrigger))
        
        with CacheRegistry.__rlock:
            try:
                (value, timestamp) = self.__cache[resource_name][key]
            except:
                return False

            if max_age is not None and not no_fault_check and \
                    (datetime.now() - timestamp).seconds > max_age:
                self.__cleanup_entry(resource_name, key, False, 
                                     cleanup_pretrigger=cleanup_pretrigger)
                return False

        return True

    def count(self, resource_name):

        return len(self.__cache[resource_name])

    def __cleanup_entry(self, resource_name, key, force, 
                        cleanup_pretrigger=None):

        _logger.debug("Doing clean-up for resource_name [%s] and key "
                      "[%s]." % (resource_name, key))

        if cleanup_pretrigger is not None:
            _logger.debug("Running pre-cleanup trigger for resource_name "
                          "[%s] and key [%s]." % (resource_name, key))

            cleanup_pretrigger(resource_name, key, force)

        del self.__cache[resource_name][key]



import logging
import resource
import re
import os
import tempfile
import shutil
import threading

import fuse

from errno import *

from gdrivefs.conf import Conf
from gdrivefs.errors import ExportFormatError, GdNotFoundError
from gdrivefs.gdfs.fsutility import dec_hint, split_path, build_filepath
from gdrivefs.gdfs.displaced_file import DisplacedFile
from gdrivefs.cache.volume import PathRelations, EntryCache, path_resolver, \
                                  CLAUSE_ID, CLAUSE_ENTRY
from gdrivefs.gdtool.drive import get_gdrive
from gdrivefs.general.buffer_segments import BufferSegments

_LOGGER = logging.getLogger(__name__)

# TODO(dustin): LCM runs in a greenlet pool. When we open a file that needs the
#               existing data for a file (read, append), a switch is done to an
#               LCM worker. If the data is absent or faulted, download the
#               content. Then, switch back.


class _OpenedManager(object):
    """Manages all of the currently-open files."""

    __opened_lock = threading.RLock()
    __fh_counter = 1

    def __init__(self):
        self.__opened = {}
        self.__opened_byfile = {}
        self.__counter = 0

        self.__temp_path = tempfile.mkdtemp()
        _LOGGER.debug("Opened-file working directory: [%s]", self.__temp_path)

    def __del__(self):
        shutil.rmtree(self.__temp_path)

    def __get_max_handles(self):

        return resource.getrlimit(resource.RLIMIT_NOFILE)[0]

    def get_new_handle(self):
        """Get a handle for a file that's about to be opened. Note that the 
        handles start at (1), so there are a lot of "+ 1" occurrences below.
        """

        cls = self.__class__
        max_handles = self.__get_max_handles()

        self.__counter += 1

        with cls.__opened_lock:
            if len(self.__opened) >= (max_handles + 1):
                raise fuse.FuseOSError(EMFILE)

            safety_counter = max_handles
            while safety_counter >= 1:
                cls.__fh_counter += 1

                if cls.__fh_counter >= (max_handles + 1):
                    cls.__fh_counter = 1

                if cls.__fh_counter not in self.__opened:
                    _LOGGER.debug("Assigning file-handle (%d).",
                                  cls.__fh_counter)

                    return cls.__fh_counter
                
        message = "Could not allocate new file handle. Safety breach."
        _LOGGER.error(message)
        raise Exception(message)

    def add(self, opened_file, fh=None):
        """Registered an OpenedFile object."""

        cls = self.__class__

        assert issubclass(opened_file.__class__, OpenedFile) is True, \
               "Can only register an OpenedFile as an opened-file."

        with cls.__opened_lock:
            if not fh:
                fh = self.get_new_handle()

            elif fh in self.__opened:
                message = ("Opened-file with file-handle (%d) has already been"
                           " registered." % (opened_file.fh))

                _LOGGER.error(message)
                raise Exception(message)

            self.__opened[fh] = opened_file

            file_path = opened_file.file_path
            if file_path in self.__opened_byfile:
                self.__opened_byfile[file_path].append(fh)
            else:
                self.__opened_byfile[file_path] = [fh]

            return fh

    def remove_by_fh(self, fh):
        """Remove an opened-file, by the handle."""

        cls = self.__class__

        with cls.__opened_lock:
            _LOGGER.debug("Closing opened-file with handle (%d).", fh)

            file_path = self.__opened[fh].file_path
            del self.__opened[fh]
            
            try:
                self.__opened_byfile[file_path].remove(fh)
            except ValueError:
                raise ValueError("Could not remove handle (%d) from list of "
                                 "open-handles for file-path [%s]: %s" % 
                                 (fh, file_path, 
                                  self.__opened_byfile[file_path]))

            if not self.__opened_byfile[file_path]:
                del self.__opened_byfile[file_path]

    def remove_by_filepath(self, file_path):

        cls = self.__class__

        _LOGGER.debug("Removing all open handles for file-path [%s].",
                      file_path)

        count = 0

        with cls.__opened_lock:
            try:
                for fh in self.__opened_byfile[file_path]:
                    self.remove_by_fh(fh)
                    count += 1
            except KeyError:
                pass

        _LOGGER.debug("(%d) file-handles removed for file-path [%s].",
                      count, file_path)

    def get_by_fh(self, fh):
        """Retrieve an opened-file, by the handle."""

        cls = self.__class__

        with cls.__opened_lock:
            if fh not in self.__opened:
                message = ("Opened-file with file-handle (%d) is not "
                          "registered (get_by_fh)." % (fh))

                _LOGGER.error(message)
                raise Exception(message)

            return self.__opened[fh]

    @property
    def opened_count(self):
        return self.__counter

    @property
    def temp_path(self):
        return self.__temp_path

_OPENED_ENTRIES_LOCK = threading.Lock()
_OPENED_ENTRIES = set()


class OpenedFile(object):
    """This class describes a single open file, and manages changes."""

    def __init__(self, entry_id, path, filename, is_hidden, mime_type):
# TODO(dustin): Until we can gracely orchestrate concurrent handles on the same 
#               entry, we can't allow it. This is referenced, just below.
        with _OPENED_ENTRIES_LOCK:
            assert entry_id not in _OPENED_ENTRIES, \
                   "Access to the same file from multiple file-handles is "\
                   "not currently supported."

            _OPENED_ENTRIES.add(entry_id)

        _LOGGER.info("Opened-file object created for entry-ID [%s] and path "
                     "(%s).", entry_id, path)

        self.__entry_id = entry_id
        self.__path = path
        self.__filename = filename
        self.__is_hidden = is_hidden
        
        self.__mime_type = mime_type
        self.__cache = EntryCache.get_instance().cache

        self.__is_loaded = False
        self.__is_dirty = False

        # Use the monotonically incremented `opened_count` to produce a unique 
        # temporary filepath.

        om = get_om()
        self.__temp_filepath = \
            os.path.join(om.temp_path, str(om.opened_count))

        self.__fh = None

        # We need to load this up-front. Since we can't do partial updates, we 
        # have to keep one whole, local copy, apply updates to it, and then 
        # post it on flush.
# TODO(dustin): Until we finish working on the download-agent so that we can 
#               have a way to orchestrate concurrent handles on the same file, 
#               we'll just have to accept the fact that concurrent access will 
#               require multiple downloads of the same file to multiple 
#               temporary files (one for each).
        self.__load_base_from_remote()

    def __del__(self):
        """This handle is being closed. Notice that we don't flush here because 
        we expect that the VFS will.
        """

        if self.__fh is not None:
            _LOGGER.debug("Removing temporary file [%s] ([%s]).", 
                          self.__temp_filepath, self.file_path)

            self.__fh.close()
            os.unlink(self.__temp_filepath)

        with _OPENED_ENTRIES_LOCK:
            _OPENED_ENTRIES.remove(self.__entry_id)

    def __repr__(self):
        replacements = { 
            'entry_id': self.__entry_id, 
            'filename': self.__filename, 
            'mime_type': self.__mime_type, 
            'is_loaded': self.__is_loaded, 
            'is_dirty': self.__is_dirty
        }

        return ("<OF [%(entry_id)s] F=[%(filename)s] MIME=[%(mime_type)s] "
                "LOADED=[%(is_loaded)s] DIRTY= [%(is_dirty)s]>" % replacements)

# TODO: We should be able to safely assume that we won't get a change event for 
#       a file until its been entirely updated, online. Therefore, the change 
#       processor should checkin, here, and make sure that any handles are 
#       closed for changed files.
#
#       We should also make sure to remove temporary file-paths in the OM temp-
#       path (if one exists) if we get a "delete" change.

    def __load_base_from_remote(self):
        """Download the data for the entry that we represent. This is probably 
        a file, but could also be a stub for -any- entry.
        """

        # If it's loaded and not-changed, don't do anything.
        if self.__is_loaded is True and self.__is_dirty is False:
            _LOGGER.debug("Not syncing-down non-dirty file.")
            return

        if self.__fh is not None:
            self.__fh.close()
            self.__fh = None

        entry = self.__cache.get(self.__entry_id)

        _LOGGER.debug("Ensuring local availability of [%s]: [%s]", 
                      entry, self.__temp_filepath)

        # Get the current version of the write-cache file, or note that we 
        # don't have it.

        _LOGGER.info("Attempting local cache update of file [%s] for entry "
                     "[%s] and mime-type [%s].",
                     self.__temp_filepath, entry, self.mime_type)

        if entry.requires_mimetype:
            length = DisplacedFile.file_size

            d = DisplacedFile(entry)
            stub_data = d.deposit_file(self.mime_type)

            self.__fh = open(self.__temp_filepath, 'w+')
            self.__fh.write(stub_data)
        else:
            _LOGGER.debug("Executing the download: [%s] => [%s]", 
                          entry.id, self.__temp_filepath)
            
            try:
# TODO(dustin): We need to inherit a file that we might've already cached by 
#               opening.
# TODO(dustin): Any call to download_to_local should use a local, temporarily 
#               file is already established. We can't use it in the reverse 
#               order though: It's one thing to already have a cache from 
#               having opened it, and it's a another thing to maintain a cache 
#               of every file that is copied.
                gd = get_gdrive()
                result = gd.download_to_local(
                            self.__temp_filepath,
                            entry,
                            self.mime_type)

                (length, cache_fault) = result
            except ExportFormatError:
                _LOGGER.exception("There was an export-format error.")
                raise fuse.FuseOSError(ENOENT)

            self.__fh = open(self.__temp_filepath, 'r+')

            self.__is_dirty = False
            self.__is_loaded = True

        _LOGGER.debug("Established base file-data for [%s]: [%s]", 
                      entry, self.__temp_filepath)

    @dec_hint(['offset', 'data'], ['data'], 'OF')
    def add_update(self, offset, data):
        """Queue an update to this file."""

        _LOGGER.debug("Applying update for offset (%d) and length (%d).",
                      offset, len(data))

        self.__is_dirty = True
        self.__fh.seek(offset)
        self.__fh.write(data)
        self.__fh.flush()

    @dec_hint(prefix='OF')
    def flush(self):
        """The OS wants to effect any changes made to the file."""

        _LOGGER.debug("Flushing opened-file.")

        entry = self.__cache.get(self.__entry_id)

        if self.__is_dirty is False:
            _LOGGER.debug("Flush will be skipped for [%s] because there "
                          "are no changes: [%s] IS_LOADED=[%s] "
                          "IS_DIRTY=[%d]", 
                          entry.id, self.file_path, self.__is_loaded, 
                          self.__is_dirty)
            return
        else:
            st = os.stat(self.__temp_filepath)

            _LOGGER.debug("Pushing (%d) bytes for entry with ID from [%s] to "
                          "GD for file-path [%s].",
                          st.st_size, entry.id, self.__temp_filepath)

# TODO: Make sure we sync the mtime to remote.
            gd = get_gdrive()
            entry = gd.update_entry(
                        entry, 
                        filename=entry.title, 
                        data_filepath=self.__temp_filepath, 
                        mime_type=self.mime_type, 
                        parents=entry.parents, 
                        is_hidden=self.__is_hidden)

            self.__is_dirty = False

# TODO(dustin): For now, we don't cleanup the temporary file. We need to 
#               schedule this using LRU-semantics.

            # Immediately update our current cached entry.

            _LOGGER.debug("Update successful. Updating local cache.")

            path_relations = PathRelations.get_instance()
            path_relations.register_entry(entry)

            _LOGGER.info("Update complete on entry with ID [%s].", entry.id)

    @dec_hint(['offset', 'length'], prefix='OF')
    def read(self, offset, length):
        
        _LOGGER.debug("Reading (%d) bytes at offset (%d).", length, offset)

        # We don't care if the cache file is dirty (not on this system, at 
        # least).

        st = os.stat(self.__temp_filepath)

        self.__fh.seek(offset)
        data = self.__fh.read(length)

        len_ = len(data)

        _LOGGER.debug("(%d) bytes retrieved from slice (%d):(%d)/(%d).",
                      len_, offset, length, st.st_size)

        if len_ != length:
            _LOGGER.warning("Read request is only returning (%d) bytes when "
                            "(%d) bytes were requested.", len_, length)

        return data

    @property
    def mime_type(self):
        return self.__mime_type

    @property
    def entry_id(self):
        return self.__entry_id

    @property
    def file_path(self):
        """The GD filepath of the requested file."""

        return build_filepath(self.__path, self.__filename)

def create_for_existing_filepath(filepath):
    """Process the file/path that was requested (potential export-type 
    directive, dot-prefix, etc..), and build an opened-file object using 
    the information.
    """

    _LOGGER.debug("Creating OpenedFile for [%s].", filepath)

    # Process/distill the requested file-path.

    try:
        result = split_path(filepath, path_resolver)
    except GdNotFoundError:
        _LOGGER.exception("Could not process [%s] (create_for_requested).",
                          filepath)

        raise fuse.FuseOSError(ENOENT)

    (parent_clause, path, filename, mime_type, is_hidden) = result
    distilled_filepath = build_filepath(path, filename)

    # Look-up the requested entry.

    path_relations = PathRelations.get_instance()

    try:
        entry_clause = path_relations.get_clause_from_path(
                        distilled_filepath)
    except:
        _LOGGER.exception("Could not try to get clause from path [%s] "
                          "(OpenedFile).", distilled_filepath)

        raise fuse.FuseOSError(EIO)

    if not entry_clause:
        _LOGGER.debug("Path [%s] does not exist for stat().", path)
        raise fuse.FuseOSError(ENOENT)

    entry = entry_clause[CLAUSE_ENTRY]

    # Normalize the mime-type by considering what's available for download. 
    # We're going to let the requests that didn't provide a mime-type fail 
    # right here. It will give us the opportunity to try a few options to 
    # get the file.

    try:
        final_mimetype = entry.normalize_download_mimetype(mime_type)
    except ExportFormatError:
        _LOGGER.exception("There was an export-format error "
                          "(create_for_requested_filesystem).")

        raise fuse.FuseOSError(ENOENT)
    except:
        _LOGGER.exception("Could not normalize mime-type [%s] for entry"
                          "[%s].", mime_type, entry)

        raise fuse.FuseOSError(EIO)

    if final_mimetype != mime_type:
        _LOGGER.info("Entry being opened will be opened as [%s] rather "
                     "than [%s].", final_mimetype, mime_type)

    # Build the object.

    return OpenedFile(
            entry_clause[CLAUSE_ID], 
            path, 
            filename, 
            is_hidden, 
            final_mimetype)

_management_instance = None
def get_om():
    global _management_instance
    if _management_instance is None:
        _management_instance = _OpenedManager()

    return _management_instance



import logging
import json
import tempfile
import os

from os import makedirs
from os.path import isdir

from gdrivefs.gdtool.drive import get_gdrive
from gdrivefs.gdtool.normal_entry import NormalEntry
from gdrivefs.conf import Conf

_logger = logging.getLogger(__name__)


class DisplacedFile(object):
    normalized_entry = None
    file_size = 1000

    def __init__(self, normalized_entry):
        assert issubclass(normalized_entry.__class__, NormalEntry) is True, \
               "DisplacedFile can not wrap a non-NormalEntry object."

        self.__normalized_entry = normalized_entry
        self.__filepath = tempfile.NamedTemporaryFile(delete=False)

    def __del__(self):
        os.unlink(self.__filepath)

    def deposit_file(self, mime_type):
        """Write the file to a temporary path, and present a stub (JSON) to the 
        user. This is the only way of getting files that don't have a 
        well-defined filesize without providing a type, ahead of time.
        """

        gd = get_gdrive()

        result = gd.download_to_local(
                    self.__filepath, 
                    self.__normalized_entry,
                    mime_type)

        (length, cache_fault) = result

        _logger.debug("Displaced entry [%s] deposited to [%s] with length "
                      "(%d).", self.__normalized_entry, self.__filepath, length)

        return self.get_stub(mime_type, length, self.__filepath)

    def get_stub(self, mime_type, file_size=0, file_path=None):
        """Return the content for an info ("stub") file."""

        if file_size == 0 and \
           self.__normalized_entry.requires_displaceable is False:
            file_size = self.__normalized_entry.file_size

        stub_data = {
                'EntryId':              self.__normalized_entry.id,
                'OriginalMimeType':     self.__normalized_entry.mime_type,
                'ExportTypes':          self.__normalized_entry.download_types,
                'Title':                self.__normalized_entry.title,
                'Labels':               self.__normalized_entry.labels,
                'FinalMimeType':        mime_type,
                'Length':               file_size,
                'RequiresMimeType':     self.__normalized_entry.requires_mimetype,
                'ImageMediaMetadata':   self.__normalized_entry.image_media_metadata
            }

        if file_path:
            stub_data['FilePath'] = file_path

            result = json.dumps(stub_data)
            padding = (' ' * (self.file_size - len(result) - 1))

            return ("%s%s\n" % (result, padding))



import logging
import stat
import dateutil.parser
import re
import json
import os
import atexit
import resource
import pprint
import math

from errno import ENOENT, EIO, ENOTDIR, ENOTEMPTY, EPERM, EEXIST
from fuse import FUSE, Operations, FuseOSError, c_statvfs, fuse_get_context, \
                 LoggingMixIn
from time import mktime, time
from sys import argv, exit, excepthook
from mimetypes import guess_type
from datetime import datetime
from os.path import split

import gdrivefs.gdfs.fsutility
import gdrivefs.gdfs.opened_file
import gdrivefs.config
import gdrivefs.config.changes
import gdrivefs.config.fs

from gdrivefs.utility import utility
from gdrivefs.change import get_change_manager
from gdrivefs.cache.volume import PathRelations, EntryCache, \
                                  CLAUSE_ENTRY, CLAUSE_PARENT, \
                                  CLAUSE_CHILDREN, CLAUSE_ID, \
                                  CLAUSE_CHILDREN_LOADED
from gdrivefs.conf import Conf
from gdrivefs.gdtool.drive import get_gdrive
from gdrivefs.gdtool.account_info import AccountInfo

from gdrivefs.gdfs.fsutility import strip_export_type, split_path,\
                                    build_filepath, dec_hint
from gdrivefs.gdfs.displaced_file import DisplacedFile
from gdrivefs.cache.volume import path_resolver
from gdrivefs.errors import GdNotFoundError
from gdrivefs.time_support import get_flat_normal_fs_time_from_epoch

_logger = logging.getLogger(__name__)

# TODO: make sure strip_extension and split_path are used when each are relevant
# TODO: make sure create path reserves a file-handle, uploads the data, and then registers the open-file with the file-handle.
# TODO: Make sure that we rely purely on the FH, whenever it is given, 
#       whereever it appears. This will be to accomodate system calls that can work either via file-path or file-handle.

def set_datetime_tz(datetime_obj, tz):
    return datetime_obj.replace(tzinfo=tz)

def get_entry_or_raise(raw_path, allow_normal_for_missing=False):
    try:
        result = split_path(raw_path, path_resolver)
        (parent_clause, path, filename, mime_type, is_hidden) = result
    except GdNotFoundError:
        _logger.exception("Could not retrieve clause for non-existent "
                          "file-path [%s] (parent does not exist)." % 
                          (raw_path))

        if allow_normal_for_missing is True:
            raise
        else:
            raise FuseOSError(ENOENT)
    except:
        _logger.exception("Could not process file-path [%s]." % 
                          (raw_path))
        raise FuseOSError(EIO)

    filepath = build_filepath(path, filename)
    path_relations = PathRelations.get_instance()

    try:
        entry_clause = path_relations.get_clause_from_path(filepath)
    except GdNotFoundError:
        _logger.exception("Could not retrieve clause for non-existent "
                          "file-path [%s] (parent exists)." % 
                          (filepath))

        if allow_normal_for_missing is True:
            raise
        else:
            raise FuseOSError(ENOENT)
    except:
        _logger.exception("Could not retrieve clause for path [%s]. " %
                          (filepath))
        raise FuseOSError(EIO)

    if not entry_clause:
        if allow_normal_for_missing is True:
            raise GdNotFoundError()
        else:
            raise FuseOSError(ENOENT)

    return (entry_clause[CLAUSE_ENTRY], path, filename)


class _GdfsMixin(object):
    """The main filesystem class."""

    def __register_open_file(self, fh, path, entry_id):
        with self.fh_lock:
            self.open_files[fh] = (entry_id, path)

    def __deregister_open_file(self, fh):
        with self.fh_lock:
            file_info = self.open_files[fh]

            del self.open_files[fh]
            return file_info

    def __get_open_file(self, fh):
        with self.fh_lock:
            return self.open_files[fh]

    def __build_stat_from_entry(self, entry):
        (uid, gid, pid) = fuse_get_context()

        block_size_b = gdrivefs.config.fs.CALCULATION_BLOCK_SIZE

        if entry.is_directory:
            effective_permission = int(Conf.get('default_perm_folder'), 
                                       8)
        elif entry.editable:
            effective_permission = int(Conf.get('default_perm_file_editable'), 
                                       8)
        else:
            effective_permission = int(Conf.get(
                                            'default_perm_file_noneditable'), 
                                       8)

        stat_result = { "st_mtime": entry.modified_date_epoch, # modified time.
                        "st_ctime": entry.modified_date_epoch, # changed time.
                        "st_atime": time(),
                        "st_uid":   uid,
                        "st_gid":   gid }
        
        if entry.is_directory:
            # Per http://sourceforge.net/apps/mediawiki/fuse/index.php?title=SimpleFilesystemHowto, 
            # default size should be 4K.
# TODO(dustin): Should we just make this (0), since that's what it is?
            stat_result["st_size"] = 1024 * 4
            stat_result["st_mode"] = (stat.S_IFDIR | effective_permission)
            stat_result["st_nlink"] = 2
        else:
            stat_result["st_size"] = DisplacedFile.file_size \
                                        if entry.requires_mimetype \
                                        else entry.file_size

            stat_result["st_mode"] = (stat.S_IFREG | effective_permission)
            stat_result["st_nlink"] = 1

        stat_result["st_blocks"] = \
            int(math.ceil(float(stat_result["st_size"]) / block_size_b))
  
        return stat_result

    @dec_hint(['raw_path', 'fh'])
    def getattr(self, raw_path, fh=None):
        """Return a stat() structure."""
# TODO: Implement handle.

        (entry, path, filename) = get_entry_or_raise(raw_path)
        return self.__build_stat_from_entry(entry)

    @dec_hint(['path', 'offset'])
    def readdir(self, path, offset):
        """A generator returning one base filename at a time."""

        # We expect "offset" to always be (0).
        if offset != 0:
            _logger.warning("readdir() has been invoked for path [%s] and "
                            "non-zero offset (%d). This is not allowed.",
                            path, offset)

# TODO: Once we start working on the cache, make sure we don't make this call, 
#       constantly.

        path_relations = PathRelations.get_instance()

        try:
            entry_clause = path_relations.get_clause_from_path(path)
        except GdNotFoundError:
            _logger.exception("Could not process [%s] (readdir).")
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not get clause from path [%s] "
                              "(readdir)." % (path))
            raise FuseOSError(EIO)

        if not entry_clause:
            raise FuseOSError(ENOENT)

        try:
            entry_tuples = path_relations.get_children_entries_from_entry_id \
                            (entry_clause[CLAUSE_ID])
        except:
            _logger.exception("Could not render list of filenames under path "
                              "[%s].", path)

            raise FuseOSError(EIO)

        yield utility.translate_filename_charset('.')
        yield utility.translate_filename_charset('..')

        for (filename, entry) in entry_tuples:

            # Decorate any file that -requires- a mime-type (all files can 
            # merely accept a mime-type)
            if entry.requires_mimetype:
                filename += utility.translate_filename_charset('#')
        
            yield (filename,
                   self.__build_stat_from_entry(entry),
                   0)

    @dec_hint(['raw_path', 'length', 'offset', 'fh'])
    def read(self, raw_path, length, offset, fh):

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            opened_file = om.get_by_fh(fh)
        except:
            _logger.exception("Could not retrieve OpenedFile for handle with"
                              "ID (%d) (read).", fh)

            raise FuseOSError(EIO)

        try:
            return opened_file.read(offset, length)
        except:
            _logger.exception("Could not read data.")
            raise FuseOSError(EIO)

    @dec_hint(['filepath', 'mode'])
    def mkdir(self, filepath, mode):
        """Create the given directory."""

# TODO: Implement the "mode".

        try:
            result = split_path(filepath, path_resolver)
            (parent_clause, path, filename, mime_type, is_hidden) = result
        except GdNotFoundError:
            _logger.exception("Could not process [%s] (mkdir).", filepath)
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not split path [%s] (mkdir).", filepath)
            raise FuseOSError(EIO)

        parent_id = parent_clause[CLAUSE_ID]
        gd = get_gdrive()

        try:
            entry = gd.create_directory(
                        filename, 
                        [parent_id], 
                        is_hidden=is_hidden)
        except:
            _logger.exception("Could not create directory with name [%s] "
                              "and parent with ID [%s].",
                              filename, parent_clause[0].id)
            raise FuseOSError(EIO)

        _logger.info("Directory [%s] created as ID [%s] under parent with "
                     "ID [%s].", filepath, entry.id, parent_id)

        #parent_clause[4] = False

        path_relations = PathRelations.get_instance()

        try:
            path_relations.register_entry(entry)
        except:
            _logger.exception("Could not register new directory in cache.")
            raise FuseOSError(EIO)

# TODO: Find a way to implement or enforce 'mode'.
    def __create(self, filepath, mode=None):
        """Create a new file.
                
        We don't implement "mode" (permissions) because the model doesn't agree 
        with GD.
        """

# TODO: Fail if it already exists.

        try:
            result = split_path(filepath, path_resolver)
            (parent_clause, path, filename, mime_type, is_hidden) = result
        except GdNotFoundError:
            _logger.exception("Could not process [%s] (i-create).", filepath)
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not split path [%s] (i-create).",
                              filepath)
            raise FuseOSError(EIO)

        distilled_filepath = build_filepath(path, filename)

        # Try to guess at a mime-type, if not otherwise given.
        if mime_type is None:
            (mimetype_guess, _) = guess_type(filename, True)
            
            if mimetype_guess is not None:
                mime_type = mimetype_guess
            else:
                mime_type = Conf.get('default_mimetype')

        gd = get_gdrive()

        try:
            entry = gd.create_file(
                        filename, 
                        [parent_clause[3]], 
                        mime_type,
                        is_hidden=is_hidden)
        except:
            _logger.exception("Could not create empty file [%s] under "
                              "parent with ID [%s].",
                              filename, parent_clause[3])

            raise FuseOSError(EIO)

        path_relations = PathRelations.get_instance()

        try:
            path_relations.register_entry(entry)
        except:
            _logger.exception("Could not register created file in cache.")
            raise FuseOSError(EIO)

        _logger.info("Inner-create of [%s] completed.", distilled_filepath)

        return (entry, path, filename, mime_type)

    @dec_hint(['filepath', 'mode'])
    def create(self, raw_filepath, mode):
        """Create a new file. This always precedes a write."""

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            fh = om.get_new_handle()
        except:
            _logger.exception("Could not acquire file-handle for create of "
                              "[%s].", raw_filepath)

            raise FuseOSError(EIO)

        (entry, path, filename, mime_type) = self.__create(raw_filepath)

        try:
            opened_file = gdrivefs.gdfs.opened_file.OpenedFile(
                            entry.id, 
                            path, 
                            filename, 
                            not entry.is_visible, 
                            mime_type)
        except:
            _logger.exception("Could not create OpenedFile object for "
                              "created file.")

            raise FuseOSError(EIO)

        _logger.debug("Registering OpenedFile object with handle (%d), "
                      "path [%s], and ID [%s].", fh, raw_filepath, entry.id)

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            om.add(opened_file, fh=fh)
        except:
            _logger.exception("Could not register OpenedFile for created "
                              "file: [%s]", opened_file)

            raise FuseOSError(EIO)

        _logger.debug("File created, opened, and completely registered.")

        return fh

    @dec_hint(['filepath', 'flags'])
    def open(self, filepath, flags):
# TODO: Fail if does not exist and the mode/flags is read only.

        try:
            opened_file = gdrivefs.gdfs.opened_file.\
                            create_for_existing_filepath(filepath)
        except GdNotFoundError:
            _logger.exception("Could not create handle for requested [%s] "
                              "(open)." % (filepath))
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not create OpenedFile object for "
                                 "opened filepath [%s].", filepath)
            raise FuseOSError(EIO)

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            fh = om.add(opened_file)
        except:
            _logger.exception("Could not register OpenedFile for opened "
                              "file.")

            raise FuseOSError(EIO)

        _logger.debug("File opened.")

        return fh

    @dec_hint(['filepath', 'fh'])
    def release(self, filepath, fh):
        """Close a file."""

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            om.remove_by_fh(fh)
        except:
            _logger.exception("Could not remove OpenedFile for handle with "
                              "ID (%d) (release).", fh)

            raise FuseOSError(EIO)

    @dec_hint(['filepath', 'data', 'offset', 'fh'], ['data'])
    def write(self, filepath, data, offset, fh):
        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            opened_file = om.get_by_fh(fh=fh)
        except:
            _logger.exception("Could not get OpenedFile (write).")
            raise FuseOSError(EIO)

        try:
            opened_file.add_update(offset, data)
        except:
            _logger.exception("Could not queue file-update.")
            raise FuseOSError(EIO)

        return len(data)

    @dec_hint(['filepath', 'fh'])
    def flush(self, filepath, fh):
        
        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            opened_file = om.get_by_fh(fh=fh)
        except:
            _logger.exception("Could not get OpenedFile (flush).")
            raise FuseOSError(EIO)

        try:
            opened_file.flush()
        except:
            _logger.exception("Could not flush local updates.")
            raise FuseOSError(EIO)

    @dec_hint(['filepath'])
    def rmdir(self, filepath):
        """Remove a directory."""

        path_relations = PathRelations.get_instance()

        try:
            entry_clause = path_relations.get_clause_from_path(filepath)
        except GdNotFoundError:
            _logger.exception("Could not process [%s] (rmdir).", filepath)
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not get clause from file-path [%s] "
                              "(rmdir).", filepath)
            raise FuseOSError(EIO)

        if not entry_clause:
            _logger.error("Path [%s] does not exist for rmdir().", filepath)
            raise FuseOSError(ENOENT)

        entry_id = entry_clause[CLAUSE_ID]
        normalized_entry = entry_clause[CLAUSE_ENTRY]

        # Check if not a directory.

        if not normalized_entry.is_directory:
            _logger.error("Can not rmdir() non-directory [%s] with ID [%s].", 
                          filepath, entry_id)

            raise FuseOSError(ENOTDIR)

        # Ensure the folder is empty.

        gd = get_gdrive()

        try:
            found = gd.get_children_under_parent_id(
                        entry_id,
                        max_results=1)
        except:
            _logger.exception("Could not determine if directory to be removed "
                              "has children.", entry_id)

            raise FuseOSError(EIO)

        if found:
            raise FuseOSError(ENOTEMPTY)

        try:
            gd.remove_entry(normalized_entry)
        except (NameError):
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not remove directory [%s] with ID [%s].",
                              filepath, entry_id)

            raise FuseOSError(EIO)
# TODO: Remove from cache.

    # Not supported. Google Drive doesn't fit within this model.
    @dec_hint(['filepath', 'mode'])
    def chmod(self, filepath, mode):
        # Return successfully, or rsync might have a problem.
#        raise FuseOSError(EPERM) # Operation not permitted.
        pass

    # Not supported. Google Drive doesn't fit within this model.
    @dec_hint(['filepath', 'uid', 'gid'])
    def chown(self, filepath, uid, gid):
        # Return successfully, or rsync might have a problem.
#        raise FuseOSError(EPERM) # Operation not permitted.
        pass

    # Not supported.
    @dec_hint(['target', 'source'])
    def symlink(self, target, source):

        raise FuseOSError(EPERM)

    # Not supported.
    @dec_hint(['filepath'])
    def readlink(self, filepath):

        raise FuseOSError(EPERM)

    @dec_hint(['filepath'])
    def statfs(self, filepath):
        """Return filesystem status info (for df).

        The given file-path seems to always be '/'.

        REF: http://www.ibm.com/developerworks/linux/library/l-fuse/
        REF: http://stackoverflow.com/questions/4965355/converting-statvfs-to-percentage-free-correctly
        """

        block_size_b = gdrivefs.config.fs.CALCULATION_BLOCK_SIZE

        try:
            account_info = AccountInfo.get_instance()
            total = account_info.quota_bytes_total / block_size
            used = account_info.quota_bytes_used / block_size
            free = total - used
        except:
            _logger.exception("Could not get account-info.")
            raise FuseOSError(EIO)

        return {
            # Optimal transfer block size.
            'f_bsize': block_size,

            # Total data blocks in file system.
            'f_blocks': total,

            # Fragment size.
            'f_frsize': block_size,

            # Free blocks in filesystem.
            'f_bfree': free,

            # Free blocks avail to non-superuser.
            'f_bavail': free

            # Total file nodes in filesystem.
#            'f_files': 0,

            # Free file nodes in filesystem.
#            'f_ffree': 0,

            # Free inodes for unprivileged users.
#            'f_favail': 0
        }

    @dec_hint(['filepath_old', 'filepath_new'])
    def rename(self, filepath_old, filepath_new):
        # Make sure the old filepath exists.
        (entry, path, filename_old) = get_entry_or_raise(filepath_old)

        # At this point, decorations, the is-hidden prefix, etc.. haven't been
        # stripped.
        (path, filename_new_raw) = split(filepath_new)

        # Make sure the new filepath doesn't exist.

        try:
            get_entry_or_raise(filepath_new, True)
        except GdNotFoundError:
            pass

        gd = get_gdrive()

        try:
            entry = gd.rename(entry, filename_new_raw)
        except:
            _logger.exception("Could not update entry [%s] for rename.", entry)
            raise FuseOSError(EIO)

        # Update our knowledge of the entry.

        path_relations = PathRelations.get_instance()

        try:
            path_relations.register_entry(entry)
        except:
            _logger.exception("Could not register renamed entry: %s", entry)
            raise FuseOSError(EIO)

    @dec_hint(['filepath', 'length', 'fh'])
    def truncate(self, filepath, length, fh=None):
        if fh is not None:
            om = gdrivefs.gdfs.opened_file.get_om()

            try:
                opened_file = om.get_by_fh(fh)
            except:
                _logger.exception("Could not retrieve OpenedFile for handle "
                                  "with ID (%d) (truncate).", fh)

                raise FuseOSError(EIO)

            _logger.debug("Truncating and clearing FH: %s", opened_file)

            opened_file.reset_state()

            entry_id = opened_file.entry_id
            cache = EntryCache.get_instance().cache
            entry = cache.get(entry_id)

            opened_file.truncate(length)
        else:
            (entry, path, filename) = get_entry_or_raise(filepath)

        gd = get_gdrive()

        try:
            entry = gd.truncate(entry)
        except:
            _logger.exception("Could not truncate entry [%s].", entry)
            raise FuseOSError(EIO)

# TODO(dustin): It would be a lot quicker if we truncate our temporary file 
#               here, and make sure its mtime matches.

        # We don't need to update our internal representation of the file (just 
        # our file-handle and its related buffering).

    @dec_hint(['file_path'])
    def unlink(self, file_path):
        """Remove a file."""
# TODO: Change to simply move to "trash". Have a FUSE option to elect this
# behavior.
        path_relations = PathRelations.get_instance()

        try:
            entry_clause = path_relations.get_clause_from_path(file_path)
        except GdNotFoundError:
            _logger.exception("Could not process [%s] (unlink).", file_path)
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not get clause from file-path [%s] "
                              "(unlink).", file_path)

            raise FuseOSError(EIO)

        if not entry_clause:
            _logger.error("Path [%s] does not exist for unlink().",
                          file_path)

            raise FuseOSError(ENOENT)

        entry_id = entry_clause[CLAUSE_ID]
        normalized_entry = entry_clause[CLAUSE_ENTRY]

        # Check if a directory.

        if normalized_entry.is_directory:
            _logger.error("Can not unlink() directory [%s] with ID [%s]. "
                          "Must be file.", file_path, entry_id)

            raise FuseOSError(errno.EISDIR)

        # Remove online. Complements local removal (if not found locally, a 
        # follow-up request checks online).

        gd = get_gdrive()

        try:
            gd.remove_entry(normalized_entry)
        except NameError:
            raise FuseOSError(ENOENT)
        except:
            _logger.exception("Could not remove file [%s] with ID [%s].",
                              file_path, entry_id)

            raise FuseOSError(EIO)

        # Remove from cache. Will no longer be able to be found, locally.
        PathRelations.get_instance().remove_entry_all(entry_id)

        # Remove from among opened-files.

        om = gdrivefs.gdfs.opened_file.get_om()

        try:
            opened_file = om.remove_by_filepath(file_path)
        except:
            _logger.exception("There was an error while removing all "
                                 "opened-file instances for file [%s] "
                                 "(remove).", file_path)
            raise FuseOSError(EIO)

    @dec_hint(['raw_path', 'times'])
    def utimens(self, raw_path, times=None):
        """Set the file times."""

        if times is not None:
            (atime, mtime) = times
        else:
            now = time()
            (atime, mtime) = (now, now)

        (entry, path, filename) = get_entry_or_raise(raw_path)

        mtime_phrase = get_flat_normal_fs_time_from_epoch(mtime)
        atime_phrase = get_flat_normal_fs_time_from_epoch(atime)

        gd = get_gdrive()

        try:
            entry = gd.update_entry(
                        entry, 
                        modified_datetime=mtime_phrase,
                        accessed_datetime=atime_phrase)
        except:
            _logger.exception("Could not update entry [%s] for times.",
                              entry)

            raise FuseOSError(EIO)

        return 0

    @dec_hint(['path'])
    def init(self, path):
        """Called on filesystem mount. Path is always /."""

        if gdrivefs.config.changes.MONITOR_CHANGES is True:
            _logger.info("Activating change-monitor.")
            get_change_manager().mount_init()
        else:
            _logger.warning("We were told not to monitor changes.")

    @dec_hint(['path'])
    def destroy(self, path):
        """Called on filesystem destruction. Path is always /."""

        if gdrivefs.config.changes.MONITOR_CHANGES is True:
            _logger.info("Stopping change-monitor.")
            get_change_manager().mount_destroy()

    @dec_hint(['path'])
    def listxattr(self, raw_path):
        (entry, path, filename) = get_entry_or_raise(raw_path)

        return entry.xattr_data.keys()

    @dec_hint(['path', 'name', 'position'])
    def getxattr(self, raw_path, name, position=0):
        (entry, path, filename) = get_entry_or_raise(raw_path)

        try:
            return entry.xattr_data[name] + "\n"
        except:
            return ''

if gdrivefs.config.DO_LOG_FUSE_MESSAGES is True:
    class GDriveFS(_GdfsMixin, LoggingMixIn, Operations):
        pass
else:
    class GDriveFS(_GdfsMixin, Operations):
        pass

def mount(auth_storage_filepath, mountpoint, debug=None, nothreads=None, 
          option_string=None):

    if os.path.exists(auth_storage_filepath) is False:
        raise ValueError("Credential path is not valid: [%s]" %
                         (auth_storage_filepath,))

    fuse_opts = {}
    
    if option_string:
        for opt_parts in [opt.split('=', 1) \
                          for opt \
                          in option_string.split(',') ]:
            k = opt_parts[0]

            # We need to present a bool type for on/off flags. Since all we
            # have are strings, we'll convert anything with a 'True' or 'False'
            # to a bool, or anything with just a key to True.
            if len(opt_parts) == 2:
                v = opt_parts[1]
                v_lower = v.lower()

                if v_lower == 'true':
                    v = True
                elif v_lower == 'false':
                    v = False
            else:
                v = True

            # We have a list of provided options. See which match against our 
            # application options.

            _logger.debug("Setting option [%s] to [%s].", k, v)

            try:
                Conf.set(k, v)
            except KeyError as e:
                _logger.debug("Forwarding option [%s] with value [%s] to "
                              "FUSE.", k, v)

                fuse_opts[k] = v

    if gdrivefs.config.IS_DEBUG is True:
        _logger.debug("FUSE options:\n%s", pprint.pformat(fuse_opts))

    _logger.debug("PERMS: F=%s E=%s NE=%s",
                  Conf.get('default_perm_folder'), 
                  Conf.get('default_perm_file_editable'), 
                  Conf.get('default_perm_file_noneditable'))

    # Assume that any option that wasn't an application option is a FUSE 
    # option. The Python-FUSE interface that we're using is beautiful/elegant,
    # but there's no help support. The user is just going to have to know the
    # options.

    set_auth_cache_filepath(auth_storage_filepath)

    # How we'll appear in diskfree, mtab, etc..
    name = ("gdfs(%s)" % (auth_storage_filepath,))

    # Make sure we can connect.
    gdrivefs.gdtool.account_info.AccountInfo().get_data()

    fuse = FUSE(
            GDriveFS(), 
            mountpoint, 
            debug=debug, 
            foreground=debug, 
            nothreads=nothreads, 
            fsname=name, 
            **fuse_opts)

def set_auth_cache_filepath(auth_storage_filepath):
    auth_storage_filepath = os.path.abspath(auth_storage_filepath)

    Conf.set('auth_cache_filepath', auth_storage_filepath)



import logging

_logger = logging.getLogger(__name__)


class LiveReaderBase(object):
    """A base object for data that can be retrieved on demand."""

    def __init__(self):
        self.__data = None

    def __getitem__(self, key):
        child_name = self.__class__.__name__

        try:
            return self.__data[key]
        except:
            pass

        self.__data = self.get_data()
        return self.__data[key]

    def get_data(self):
        raise NotImplementedError("get_data() method must be implemented in "
                                  "the LiveReaderBase child.")

    @classmethod
    def get_instance(cls):
        """A helper method to dispense a singleton of whomever is inheriting "
        from us.
        """

        class_name = cls.__name__

        try:
            LiveReaderBase.__instances
        except:
            LiveReaderBase.__instances = {}

        try:
            return LiveReaderBase.__instances[class_name]
        except:
            LiveReaderBase.__instances[class_name] = cls()
            return LiveReaderBase.__instances[class_name]



#!/usr/bin/env python2.7

import os.path
import setuptools

import gdrivefs

_APP_PATH = os.path.dirname(gdrivefs.__file__)

with open(os.path.join(_APP_PATH, 'resources', 'README.rst')) as f:
      long_description = f.read()

with open(os.path.join(_APP_PATH, 'resources', 'requirements.txt')) as f:
      install_requires = [s.strip() for s in f.readlines()]

setuptools.setup(
    name='gdrivefs',
    version=gdrivefs.__version__,
    description="A complete FUSE adapter for Google Drive.",
    long_description=long_description,
    classifiers=[
        'Topic :: System :: Filesystems',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Internet',
        'Topic :: Utilities'
    ],
    keywords='google-drive google drive fuse filesystem',
    author='Dustin Oprea',
    author_email='myselfasunder@gmail.com',
    url='https://github.com/dsoprea/GDriveFS',
    license='GPL 2',
    packages=setuptools.find_packages(exclude=['dev', 'tests']),
    include_package_data=True,
    package_data={
        'gdrivefs': [
            'resources/README.rst',
            'resources/requirements.txt',
        ],
    },
    zip_safe=False,
    install_requires=install_requires,
    scripts=[
        'gdrivefs/resources/scripts/gdfs',
        'gdrivefs/resources/scripts/gdfstool',
        'gdrivefs/resources/scripts/gdfsdumpentry',
    ],
)
